//
// FluXX   - Photometric Color Calibration
//
// Copyright Hartmut V. Bornemann, 2017, 2019
//
// This script contains the DSMAngle class from:
//
//   Copyright (C) 2012-2019, Andres del Pozo
//   All rights reserved.
//
//
// This is a new version based on a windows program FluXX.exe, developed early
// 2017. The development was gladly supported by the PixInsight Austria team:
//    Herbert Walter
//    Dr. Franz Gruber
//    Gerald Wechselberger
//
// History
//
// date        version  reason
//
// 16.04.21    3.2.0    ImageSolver removed, Gaia local catalog included
// 16.05.20    3.1.10   adopted to PixInsight 1.8.8-7
// 16.05.20    3.1.8    catalog box creation wrong with rotated images
// 27.09.19    3.1.7    Gmag in SDSS corrected
// 25.06.19    3.1.6    max magnitude added to GUI
// 22.05.19    3.2.0    transformFluxes, zero magnitudes Gaia
// 19.03.19    3.2.0    VOTABLE
// 19.03.19    3.1.2    Gaia photometric calibration
// 10.03.19    3.1.1    bug, astrometric solution
// 03.03.19    3.1.1    bug, astrometric catalog preset
// 02.03.19    3.1.1    bug, if image not solved
// 26.02.19    3.1.0    integrated plate solver
// 18.02.19    3.0.3    star count limit
// 15.02.19    3.0.2    new help text format p2i (pdf to image)
// 13.02.19    3.0.1    file error, if Gaia passband data file not exists
//                      a warning message is written to the console
// 12.02.19    3.0.0    release

#include <pjsr/StarDetector.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/Sizer.jsh>
#include <pjsr/Slider.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/SampleType.jsh>
#include <pjsr/DataType.jsh>
#include <pjsr/Color.jsh>
#include <pjsr/FileMode.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/StdDialogCode.jsh>

#feature-id    Utilities > FluXX
#define ID     "FluXX"
#define TITLE  "Photometric Color Calibration"
#define VERSION "3.2.0"


#define DEFAULT_AUTOSTRETCH_SCLIP  -2.80
#define DEFAULT_AUTOSTRETCH_TBGND   0.25

#define GAIA   "GAIA-DR2"
#define GAIA3  "GAIA-DR3"
#define SDSS   "SDSS-DR9"

#define PDFDOC    "/doc/scripts/" + ID + "/FluXX.p2i"

#define ColorSpace_HSI 6

#define ProcessExitStatus_Normal    0 // The process finished normally
#define ProcessExitStatus_Crash     1 // The process crashed


var sites = ["http://vizier.u-strasbg.fr/",           // (cds) (fr)
             "http://vizier.cfa.harvard.edu/",        // (cfa) (us)
             "http://vizier.hia.nrc.ca/",             // (cadc) (ca)
             "http://vizier.nao.ac.jp/",              // (adac) (jp)
             "http://VizieR.china-vo.org/",           // (bejing) (cn)
             "http://vizier.ast.cam.ac.uk/",          // (cambridge) (uk)
             "http://vizier.inasan.ru/",              // (moscow) (ru)
             "http://www.ukirt.jach.hawaii.edu/",     // (ukirt) (hawaii)
             "http://viziersaao.chpc.ac.za/"
            ];

var catalogTypes =
   [
      {id:"APASS",
      sid:"APASS",
      queryColumns:["RAJ2000", "DEJ2000", "Bmag", "Vmag", "g\'mag", "r\'mag", "i\'mag", "nobs"],
      dataColumns: ["RAJ2000", "DEJ2000", "Bmag", "Vmag", "g\'mag", "r\'mag", "i\'mag", "nobs"],
      properties:  ["ra", "dec", "Bmag", "Vmag", "g_mag", "r_mag", "i_mag", "nobs"],
      tableName:"II/336/apass9",
      criteriaMagName:"Vmag"},

      {id:GAIA,
      sid:"GAIA",
      queryColumns:["RAJ2000", "DEJ2000", "Gmag", "BPmag", "RPmag"],
      dataColumns: ["RAJ2000", "DEJ2000", "Gmag", "BPmag", "RPmag"],
      properties:  ["ra", "dec", "Gmag", "BPmag", "RPmag"],
      tableName: "I/345/gaia2",
      criteriaMagName:"Gmag"},

      {id:GAIA3,
      sid:"GAIA3",
      queryColumns:["RAJ2000", "DEJ2000", "Gmag", "BPmag", "RPmag"],
      dataColumns: ["RAJ2000", "DEJ2000", "Gmag", "BPmag", "RPmag"],
      properties:  ["ra", "dec", "Gmag", "BPmag", "RPmag"],
      tableName: "local",
      criteriaMagName:"Gmag"},

      {id:SDSS,
      sid:"SDSS",
      queryColumns:["_RA", "_DE", "gmag", "rmag", "cl"],
      dataColumns: ["_RAJ2000", "_DEJ2000", "gmag", "rmag", "cl"],
      properties:  ["ra", "dec", "gmag", "rmag", "cl"],
      tableName:"V/139/sdss9",
      criteriaMagName:"gmag"}
   ];

function gccParms()
{
   this.minMag          = 7.0;
   this.maxMag          = 15.0;
   this.autoMagnitude   = true;
   this.magnitude       = 12;
   this.maxStars        = 500;   // maximum number of stars for the evaluation process
   this.numStars        = 1000;  // maximum number of stars for each download segment
                                 // or Gaia catalog read
   this.allStars        = true;
   this.sensitivity     = 0;
   this.catalogName     = GAIA;     // full name - release
   this.tablename       = "";
   this.numObs          = 2;        // APASS #of observations
   this.plot            = false;
   this.createPreview   = false;
   this.VizierSite      = "";
   this.focalLength     = 1000;
   this.pixelSize       = 9;
   this.catalogMode     = 0;        // 0=local, 1=remote, 2=auto
   this.localCatalog    = '';       // local catalog
   this.catalogSolver   = '';
   this.workingDirectory = File.systemTempDirectory;
   this.aggregat        = '';
   this.boxes           = [];
   this.lowerLimit      = 0;
   this.upperLimit      = 0.1;
   this.target          = 0.001;
}

function cControl()
{
   this.objectName      = "";
   this.catalogType     = null;
   this.width           = 0;
   this.height          = 0;
   this.radius          = 0;
   this.ra              = 0;
   this.dec             = 0;
   this.dateobs         = new Date();
   this.epoch           = 0;
   this.resolution      = 1;
   this.rotation        = 1;
   this.frame           = null;
   this.localCatValid   = false;
   this.catalogValid    = false;
   this.reScan          = true;
   this.joined          = false;
   this.filterSet       = true;
   this.magnitudesSet   = true;
   this.plotted         = false;
   this.success         = false;
   this.windowRG        = null;
   this.windowRB        = null;
   this.reloadCatalog   = false;
   this.preView         = null;
   this.window          = null;
   this.circleFactor    = 1;     // factor area circle/ area image rectangle
   this.imageList       = null;
   this.helpTextVisible = true;
   this.page            = 0;
   this.PixInsightRoot  = "";   // C:\Program Files\PixInsight
   this.gnuplotExe      = "";
}

function infoDialog(controlParms)
{
   this.__base__ = Dialog;
   this.__base__();

   var dlg = this;

   var cc = controlParms;
   //
   // FluXX.p2i
   //
   var filename = cc.PixInsightRoot + PDFDOC;

   var bmp  = null;
   var scl  = null;
   var imgCount = 0;

   this.onKeyPress = function( keyCode, modifiers )
   {
      if (cc.imageList == null) return;
      if (keyCode == 268435460 | keyCode == 268436480) nextPage();
      if (keyCode == 268435458 | keyCode == 268436224) prevPage();
   }

   this.onMouseWheel = function( x, y, delta, buttonState, modifiers )
   {
      if (delta > 0) prevPage();
      if (delta < 0) nextPage();
   }

   this.helpLabel = new Label( this );
   with ( this.helpLabel )
   {
      frameStyle = FrameStyle_Box;
      margin = 4;
      wordWrapping = true;
      useRichText = true;
      visible = cc.helpTextVisible;

      text = "<p><b>FluXX " + TITLE + " v" + VERSION + "</b></p>" +
      "<p><b>Purpose</p>" +
      "Calculation of RGB color calibration factors for the use in PixInsight ColorCalibration" +
      " process. FluXX locates stars in the image frames and matches their flux counts with the " +
      "photometric catalogs<p>" +
      "<p><b>Supporters</p>" +
      "<p>With thanks for beta testing and publishing to Dr. Franz Gruber, Herbert Walter, Gerald Wechselberger</p>" +
      "<p><b>PixInsight</p>" +"<p>Plate solving of astronomical images.</p>" +
      "<p>Copyright (C) 2012-2019, Andres del Pozo</p>" +
      "<p>All rights reserved.</p>" +
      "<p><b>Simbad</p>" +
      "<p>This research has made use of the SIMBAD database, operated at CDS, Strasbourg, France</p>" +
      "<p><b>VizieR</p>" +
      "<p>This research has made use of the VizieR catalogue access tool, CDS, Strasbourg, France</p>" +
      "<p><b>APASS</p>" +
      "<p>The AAVSO Photometric All-Sky Survey project</p>" +
      "<p>This research was made possible through the use of the AAVSO Photometric All-Sky Survey (APASS), " +
      "funded by the Robert Martin Ayers Sciences Fund and NSF AST-1412587.</p>" +
      "<p><b>GAIA</p>" +
      "<p>This work has made use of data from the European Space Agency (ESA) " +
      "mission Gaia " +
      "(https://www.cosmos.esa.int/gaia), processed by the Gaia Data Processing " +
      "and Analysis Consortium " +
       "(DPAC, https://www.cosmos.esa.int/web/gaia/dpac/consortium). " +
     "Funding for the DPAC has been provided by national institutions, " +
      "in particular the institutions participating in the Gaia Multilateral Agreement.</p>" +
      "<p><b>SDSS</p>" +
      "<p>Photometric Flux Calibration by Sload Digital Sky Survey</p>" +
      "<p>Funding for SDSS-III has been provided by the Alfred P. Sloan Foundation, " +
      "the Participating Institutions, the National Science Foundation, and the U.S. " +
      "Department of Energy Office of Science. The SDSS-III web site is " +
      "http://www.sdss3.org/.</p>" +
      "SDSS-III is managed by the Astrophysical Research Consortium for the Participating Institutions " +
      "of the SDSS-III Collaboration including the University of Arizona, the Brazilian Participation Group, " +
      "Brookhaven National Laboratory, University of Cambridge, University of Florida, the French Participation " +
      "Group, the German Participation Group, the Instituto de Astrofisica de Canarias, the Michigan State/Notre " +
      "Dame/JINA Participation Group, Johns Hopkins University, Lawrence Berkeley National Laboratory, Max Planck " +
      "Institute for Astrophysics, New Mexico State University, New York University, Ohio State University, " +
      "Pennsylvania State University, University of Portsmouth, Princeton University, the Spanish Participation " +
      "Group, University of Tokyo, University of Utah, Vanderbilt University, University of Virginia, University " +
      "of Washington, and Yale University." +
      "<p><b>mailto:hvb356@hotmail.de</p>";
   }

   this.docBox = new Frame(this);
   with (this.docBox)
   {
      visible = !cc.helpTextVisible;

      onPaint = function (x0, y0, x1, y1)
      {
         if (scl == null) return;
         var w = scl.width;
         var h = scl.height;
         var x = (x1 - w) / 2;
         var y = (y1 - h) / 2;
         var g = new Graphics( this );
         g.drawBitmap( x, y, scl );
         g.end();
         // load @ next
         if (cc.page < cc.imageList.length - 2)
         {
            if (cc.imageList[cc.page + 1] == null)
               cc.imageList[cc.page + 1] = jpgToImage(cc.page + 1);
         }
      }

      onResize = function (wNew, hNew, wOld, hOld)
      {
         if (bmp == null) return;
         scl = scaleImage(bmp, wNew, hNew);
         repaint();
      }
   }
/*
   this.btnPDF = new PushButton(this);
   with (this.btnPDF)
   {
      text = "Document";
      onPress = function()
      {
         if (cc.helpTextVisible)
         {
            //
            // flip button face
            //
            cc.helpTextVisible = !cc.helpTextVisible;
            parent.docBox.visible = !cc.helpTextVisible;
            parent.helpLabel.visible = cc.helpTextVisible;
            text = "Acknowledgements";

            if (cc.imageList == null)
            {
               //
               // extract file
               //
               var p2i     = new File(filename);
               imgCount    = p2i.read(DataType_Int32);
               cc.imageList   = new Array(imgCount);
               //
               // write jpg file(s)
               //
               for (var i = 0; i < imgCount; i++)
               {
                  var length = p2i.read(DataType_Int32);
                  var buffer = p2i.read( DataType_ByteArray, length );
                  File.writeFile(jpgName(i), buffer );
                  cc.imageList[i] = null;
               }
               bmp = loadImage(cc.page, cc.imageList);
               var vp = parent.docBox;
               scl = scaleImage(bmp, vp.width, vp.height);
               vp.repaint();
            }
         }
         else
         {
            cc.helpTextVisible = !cc.helpTextVisible;
            parent.docBox.visible = !cc.helpTextVisible;
            parent.helpLabel.visible = cc.helpTextVisible;
            text = "Document";
         }
      }
   }
   */

   this.btnOK = new PushButton(this);
   with (this.btnOK)
   {
      icon = this.scaledResource(":/icons/ok.png");
      text = "OK";
      onPress = function() parent.done(0);
   }

   var buttonsSizer = new HorizontalSizer();
   buttonsSizer.margin = 4;
   //buttonsSizer.add(this.btnPDF);
   buttonsSizer.addStretch();
   buttonsSizer.add(this.btnOK);

   this.sizer = new VerticalSizer(this);
   this.sizer.add (this.helpLabel);
   this.sizer.add (this.docBox);
   this.sizer.addSpacing(12);
   this.sizer.add(buttonsSizer);

   this.setScaledMinSize(1000, 750);

   this.windowTitle = "FluXX - Info";

	this.adjustToContents();
 	this.userResizable = true;
   processEvents();
   this.bringToFront();

   if (!cc.helpTextVisible)
   {
      cc.page -= 1;
      nextPage();
   }

   // local functions

   function nextPage()
   {
      if (cc.page >= cc.imageList.length - 1) return;
      cc.page += 1;
      bmp = loadImage(cc.page, cc.imageList);
      scl = scaleImage(bmp, dlg.docBox.width, dlg.docBox.height);
      dlg.docBox.repaint();
   }

   function prevPage()
   {
      if (cc.page <= 0) return;
      cc.page -= 1;
      bmp = loadImage(cc.page, cc.imageList);
      scl = scaleImage(bmp, dlg.docBox.width, dlg.docBox.height);
      dlg.docBox.repaint();
   }

   function jpgName(i)
   {
      return File.systemTempDirectory + '/' + ID + "[" + i.toString() + "].jpg";
   }

   function jpgToImage(index)
   {
      var jpgFile    = jpgName(index);
      var fileFormat = new FileFormat( ".jpg", true/*toRead*/, false/*toWrite*/ );
      var format     = new FileFormatInstance( fileFormat );
      var img        = new Image();
      var info       = format.open( jpgFile, "verbosity 0" );
      if ( info.length <= 0 )
         throw new Error( "Unable to open input file: " + temp );
      format.readImage(img);
      return img;
   }

   function loadImage(index, imageList)
   {
      if (index >= imageList.length) return null;
      if (imageList[index] == null)
      {
         imageList[index] = jpgToImage(index);
      }
      return imageList[index].render();
   }

   function scaleImage(bmp, width, height)
   {
      var scale = Math.min(width / bmp.width, height / bmp.height);
      var w = bmp.width * scale;
      var h = bmp.height * scale;
      return bmp.scaled(scale);
   }

}

// **************************************************************************
// **************************************************************************
// **************************************************************************
//
//
//                         FluXX dialog
//
//
// **************************************************************************
// **************************************************************************
// **************************************************************************
function showDialog(imageWindow)
{
   this.__base__ = Dialog;
   this.__base__();
   var dlg = this;

   var parms = new gccParms();

   var pNames = Object.keys(parms).join(',');
   var setts = Settings.read(ID + "/parms", DataType_String );

   // Console.writeln(setts);

   /*
   for (var i in catalogTypes)
   {
      var t = catalogTypes[i];
      Console.writeln("ID "+t.id+'\t'+t.tableName+'\t'+t.criteriaMagName);
      for (var j = 0; j < t.queryColumns.length; j++)
         Console.writeln('\t'+"Query " + t.queryColumns[j]);
      for (var j = 0; j < t.dataColumns.length; j++)
         Console.writeln('\t'+"Data  " + t.dataColumns[j]);
      for (var j = 0; j < t.properties.length; j++)
         Console.writeln('\t'+"Prop  " + t.properties[j]);
   }
   */

   var headLine = "FluXX photometric color calibration on " + imageWindow.mainView.id;
   Console.writeln(stringOf('*', headLine.length));
   Console.writeln(headLine);
   Console.writeln(stringOf('*', headLine.length));

   if (setts != null) parms = JSON.parse(setts);

   if (Object.keys(parms).join(',') != pNames)
   {
      parms = new gccParms();
   }
   if (parms.VizierSite == "") parms.VizierSite = sites[0];

   var cc    = new cControl();
   with (cc)
   {
      catalogType = getCatalogType(parms.catalogName)
      window = imageWindow;
      width  = window.currentView.image.width;
      height = window.currentView.image.height;
      frame  = new Rect(8, 8, width - 8, height - 8);
      PixInsightRoot = File.currentWorkingDirectory;;
      gnuplotExe = gnuplotExecutable();
      localCatValid = fileExists(parms.localCatalog);
      resolution = imageResolution(parms.focalLength, parms.pixelSize);
   }

   var view   = cc.window.mainView;

   cc.window.regenerateAstrometricSolution();
   cc.resolution = getImageResolution(cc.window) * 3600;   // arc sec

   var cat   = null;
   var spinBoxWidth = 7*this.font.width( 'M' );

   var starCollections  = null;
   var joinedFluxes = null;

   var Analysis     = null;

   // **************************************************************************
   // my © and id label
   //
   this.lblCopyright = new Label(this);
   this.lblCopyright.text      = "© 2019, Hartmut V. Bornemann";
   this.lblCopyright.textAlignment = TextAlign_Left | TextAlign_VertCenter;

   this.iconInfo = new ToolButton(this);
   with (this.iconInfo)
   {
      icon = this.scaledResource(":/icons/info.png");
      onPress = function()
      {
         var dialog = new infoDialog(cc);
	      dialog.execute();
         processEvents();
         dlg.bringToFront();
      }
   }

   var topLineSizer = new HorizontalSizer();
   topLineSizer.margin = 4;
   topLineSizer.add(this.lblCopyright);
   topLineSizer.addStretch();
   topLineSizer.add(this.iconInfo);

   this.lblView = new Label(this);
   this.lblView.text = "<b>View " + view.id + "</b>";
   this.lblView.useRichText = true;

   // **************************************************************************

   var labelWidth1 = this.font.width( "MMMRight Ascension:" + "M" );
   var labelWidth2 = this.font.width( "MFocal length:" + "M" );
   var editWidth1  = this.font.width( "MM+00 00 00.000" );
   var editWidth2  = this.font.width( "9999.999M" );

   this.lblRa     = new fieldLabel(this, "Right Ascension:", labelWidth1);
   this.editRa    = new Edit(this);
   this.editRa.setFixedWidth(editWidth1);
   this.editRa.rightAlignment  = true;
   this.editRa.toolTip = "<b><p>Right Ascension (J2000)</p></b>" +
      "<p>Enter hours formatted 'hh mm ss.ddd'</p>";
   this.editRa.onTextUpdated = function (text) {solved = false;}

   this.lblDec    = fieldLabel(this, "Declination:", labelWidth1);
   this.editDec   = new Edit(this);
   this.editDec.setFixedWidth(editWidth1);
   this.editDec.rightAlignment  = true;
   this.editDec.toolTip = "<b><p>Declination (J2000)</p></b>" +
      "<p>Enter degrees formatted '+|-dd mm ss.ddd'</p>";
   this.editDec.onTextUpdated = function (text) {solved = false;}

   this.lblDate   = new fieldLabel(this, "Date of Observation:", labelWidth1);

   this.editDate  = new Edit(this);
   this.editDate.setFixedWidth(editWidth1);
   this.editDate.rightAlignment  = true;
   this.editDate.toolTip = "<b><p>Observation date</p></b>" +
      "<p>Date formatted 'yyyy-mm-dd'</p>";
   this.editDate.onEditCompleted = function()
   {
      cc.dateobs = new Date(Date.parse(dlg.editDate.text));
      cc.epoch = julianDay(cc.dateobs);
   }

   this.lblFocal  = new fieldLabel(this, "Focal length:", labelWidth1);
   this.editFocal = new Edit(this);
   this.editFocal.setFixedWidth(editWidth1);
   this.editFocal.rightAlignment  = true;
   this.editFocal.text     = parms.focalLength.toFixed(3);
   this.editFocal.toolTip  = "<b><p>Resolution</p></b>" +
      "<p>Calculate image resolution from focal length and pixel size</p>";
   this.editFocal.onEditCompleted = function()
   {
      var a = this.text.tryToDouble();
      if (a != null)
      {
         parms.focalLength = a;
         cc.resolution = imageResolution(parms.focalLength, parms.pixelSize);
         dlg.editResol.text = cc.resolution.toFixed(3);
      }
   }

   this.lblPixSz  = new Label(this);
   this.lblPixSz.text = "PixelSize:";

   this.editPixSz = new Edit(this);
   this.editPixSz.setFixedWidth(editWidth2);
   this.editPixSz.rightAlignment  = true;
   this.editPixSz.text     = parms.pixelSize.toFixed(1);
   this.editPixSz.toolTip  = "<b><p>Resolution</p></b>" +
      "<p>Calculate image resolution from focal length and pixel size</p>";
   this.editPixSz.onEditCompleted = function(text)
   {
      var a = this.text.tryToDouble();
      if (a != null)
      {
         parms.pixelSize = a;
         cc.resolution = imageResolution(parms.focalLength, parms.pixelSize);
         dlg.editResol.text = cc.resolution.toFixed(3);
      }
   }

   this.lblResol  = fieldLabel(this, "Resolution:", labelWidth1);

   this.editResol = new Edit(this);
   this.editResol.setFixedWidth(editWidth2);
   this.editResol.rightAlignment  = true;
   this.editResol.text = cc.resolution.toFixed(3);
   this.editResol.toolTip = "<b><p>Resolution</p></b>" +
      "<p>Overwriding image resolution changes focal length</p>";
   this.editResol.onEditCompleted = function()
   {
      var a = this.text.tryToDouble();
      if (a != null)
      {
         cc.resolution = a;
         parms.focalLength  = 206.265 * parms.pixelSize / cc.resolution;
         dlg.editFocal.text = parms.focalLength.toFixed(3);
      }
   }

   this.rbAuto    = new RadioButton(this);
   this.rbAuto.text = "Automatic";
   this.rbAuto.checked = parms.catalogMode == 2;
   this.rbAuto.toolTip = "<b><p>Plate solver catalog</p></b>" +
      "<p>Let plate solver choose a catalog</p>";
   this.rbAuto.onCheck = function(checked)
   {
      if (checked) parms.catalogMode = 2;
   }

   this.rbLocal    = new RadioButton(this);
   with (this.rbLocal)
   {
      text = "Local";
      toolTip = "<b><p>Plate solver catalog</p></b>" +
         "<p>Use a local catalog to solve the image. " +
         "The local catalog is fast and independent from web access</p>";
      enabled = cc.localCatValid;
      if (enabled)
      {
         toolTip += "<p>Current: " + parms.localCatalog + "</p";
         if (parms.catalogMode == 0) checked = true;
      }
      else
      {
         parms.catalogMode = 1;
         dlg.rbAuto.checked = true;
      }
      onCheck = function(checked)
      {
         if (checked) parms.catalogMode = 0;
      }
   }

   this.btnLoadLocalCat = new ToolButton(this);
   with (this.btnLoadLocalCat)
   {
      icon = this.scaledResource( ":/icons/select-file.png" );
      setScaledFixedSize( 20, 20 );
      toolTip = "<p>Select a local catalog file.</p>";
      onClick = function(checked )
      {
         var ofd = new OpenFileDialog();
         ofd.caption = "Select the local catalog path";
         ofd.filters = [["All supported catalog files", "*.bin,*.txt"],
                        ["Star database files", "*.bin"],
                        ["Custom catalog files", "*.txt"]];
         if (ofd.execute())
         {
            parms.localCatalog  = ofd.fileName;
            cc.localCatValid    = true;
            dlg.rbLocal.enabled = true;
            enabled = true;
         }
      }
   }

   this.rbRemote    = new RadioButton(this);
   this.rbRemote.text = "Vizier";
   this.rbRemote.checked = parms.catalogMode == 1;
   this.rbRemote.toolTip = "<b><p>Plate solver catalog</p></b>" +
      "<p>Select a catalog hosted on a server to solve the image. " +
      "This option should be taken for specific demands</p>";
   this.rbRemote.onCheck = function(checked)
   {
      if (checked) parms.catalogMode = 1;
   }

   this.lblCat    = new Label(this);
   this.lblCat.textAlignment = TextAlign_Left|TextAlign_VertCenter;
   this.lblCat.text = "Catalog:";

/*
   this.astrCat   = new ComboBox(this);
   with (this.astrCat)
   {
      editEnabled = false;
      var j = -1;
      for (var i in solver.solverCfg.availableCatalogs)
      {
         var catName = solver.solverCfg.availableCatalogs[i].name;
         addItem(catName);
         if (parms.catalogSolver == '')
         {
            parms.catalogSolver = catName;
         }
         if (parms.catalogSolver == catName) j = i;
      }
      currentItem = Math.floor(j);
      onItemSelected = function ()
      {
         parms.catalogSolver = solver.solverCfg.availableCatalogs[dlg.astrCat.currentItem].name;
         Console.writeln(parms.catalogSolver + " selected");
      }
    }
    */

   // Magnitude

   this.magnitude_Label = new fieldLabel( this, "Limit magnitude:", labelWidth1 );

   this.magnitude_SpinBox = new SpinBox( this );
   this.magnitude_SpinBox.minValue = 0;
   this.magnitude_SpinBox.maxValue = 30;
   this.magnitude_SpinBox.value = parms.magnitude;
   this.magnitude_SpinBox.toolTip = "<p>Maximum star magnitude to use for the image registration and plate-solving algorithms.<br/>" +
                           "For wider fields, use lower values.</p>";
   this.magnitude_SpinBox.setFixedWidth( spinBoxWidth );
   this.magnitude_SpinBox.enabled = !parms.autoMagnitude;
   this.magnitude_SpinBox.onValueUpdated = function( value )
   {
      parms.magnitude = value;
   };

   this.autoMagnitude_Check = new CheckBox(this);
   this.autoMagnitude_Check.text = "Automatic magnitude";
   this.autoMagnitude_Check.toolTip = "<p>The script automatically selects the optimal magnitude limit based on the estimated field of view of the image.</p>";
   this.autoMagnitude_Check.checked = parms.autoMagnitude;
   this.autoMagnitude_Check.onCheck = function (checked)
   {
      parms.autoMagnitude = checked;
      this.dialog.magnitude_SpinBox.enabled = !checked;
   };


   var magnitude_Sizer = new HorizontalSizer;
   magnitude_Sizer.spacing = 4;
   magnitude_Sizer.add(this.magnitude_Label);
   magnitude_Sizer.addSpacing(4);
   magnitude_Sizer.add(this.magnitude_SpinBox);
   magnitude_Sizer.addSpacing(4);
   magnitude_Sizer.add(this.autoMagnitude_Check);


   this.objName   = new Label(this);
   this.objName.text= "Object name:";
   this.objEdit   = new Edit(this);
   this.objEdit.onEditCompleted = function() {dlg.btnSesame.hasFocus = true;}

   this.btnSesame = new PushButton(this);
   with (this.btnSesame)
   {
      icon     = scaledResource( ":/icons/find.png" );
      text     = "Search";
      toolTip  = "<b><p>Query Sesame name resolver</p></b><p>" +
         "Get name and celestial coordinates</p>";
      onClick = function(checked )
      {
         var objectName = dlg.objEdit.text;
         while (objectName.contains(' ')) {objectName = objectName.replace(' ', '%20')};

		   var server = "http://cdsweb.u-strasbg.fr";
         var url = server + "/cgi-bin/nph-sesame/-o?" + objectName;
         var filePath = File.systemTempDirectory + "/sesameQueryResult.txt";
         var sesame = new FileDownload( url, filePath );
         if (sesame.perform())
         {
            var lines = File.readTextFile( filePath );
            Console.writeln(lines);
            lines = lines.split('\n');
            var coords = false;
            var objName = false;
            for (var i in lines)
            {
               if (lines[i].startsWith("%J "))
               {
                  var a = lines[i].split(' ');
                  var ra = a[1].tryToDouble();
                  var de = a[2].tryToDouble();
                  var strRa = DMSangle.FromAngle(ra / 15).ToString(true);
                  var strDe = DMSangle.FromAngle(de).ToString();
                  dlg.editRa.text  = strRa;
                  dlg.editDec.text = strDe;
                  coords = true;
               }
               if (lines[i].startsWith("%I.0 "))
               {
                  objectName = lines[i].substring(5).trim();
                  objName = true;
               }
               if(coords && objName) break;
            }
            if (objName)
               cc.objectName = objectName;
            else
            {
               objectName = dlg.objEdit.text.trim().toUpperCase();
               while (objectName.contains(' ')) objectName = objectName.replace(' ', '');

               var s1 = "";
               var j = -1;
               var chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
               for (var i = 0; i < objectName.length; i++)
               {
                  var c = objectName.charAt(i);
                  if(chars.contains(c))
                  {
                     s1 += c;
                  }
                  else
                  {
                     j = i;
                     break;
                  }
               }
               var o = s1;
               if (j > -1) o += ' ' + objectName.substring(j);
               objectName = o;

               cc.objectName = objectName;
            }
            Console.writeln("Object: " + cc.objectName);
            dlg.objEdit.text = cc.objectName;
         }
         else
            Console.writeln("No result from: " + url);
      }
   }

   this.btnClearSolution = new PushButton(this);
   this.btnClearSolution.text = "Clear solution";
   this.btnClearSolution.visible = false;

   this.astrometryBox = new GroupBox(this);
   with (this.astrometryBox)
   {
      visible = false;
      title = "Astrometry";
      var leftBox = new VerticalSizer();
      var rightBox = new VerticalSizer();

      var lineRa   = new HorizontalSizer();
      var lineDec  = new HorizontalSizer();
      var lineDate = new HorizontalSizer();
      var lineFocl = new HorizontalSizer();
      var lineReso = new HorizontalSizer();
      var lineDate = new HorizontalSizer();
      var lineCat  = new HorizontalSizer();

      lineRa.add(this.lblRa);
      lineRa.addSpacing(8);
      lineRa.add(this.editRa);
      lineRa.addStretch();

      lineDec.add(this.lblDec);
      lineDec.addSpacing(8);
      lineDec.add(this.editDec);
      lineDec.addStretch();

      lineDate.add(this.lblDate);
      lineDate.addSpacing(8);
      lineDate.add(this.editDate);
      lineDate.addStretch();

      lineFocl.add(this.lblFocal);
      lineFocl.addSpacing(8);
      lineFocl.add(this.editFocal);
      lineFocl.addStretch();
      lineFocl.add(this.lblPixSz);
      lineFocl.addSpacing(8);
      lineFocl.add(this.editPixSz);

      lineReso.add(this.lblResol);
      lineReso.addSpacing(8);
      lineReso.add(this.editResol);
      lineReso.addStretch();

      lineCat.add(this.lblCat);
      lineCat.addSpacing(8);
      lineCat.add(this.rbAuto);
      lineCat.addSpacing(8);
      lineCat.add(this.rbLocal);
      lineCat.addSpacing(8);
      lineCat.add(this.btnLoadLocalCat);
      lineCat.addSpacing(8);
      lineCat.add(this.rbRemote);
      lineCat.addSpacing(8);
      lineCat.addSpacing(8);
      //lineCat.add(this.astrCat);


      with (leftBox)
      {
         margin = 4;
         add(lineRa);
         addSpacing(4);
         add(lineDec);
         addSpacing(4);
         add(lineDate);
         addSpacing(4);
         add(lineFocl);
         addSpacing(4);
         add(lineReso);
         addSpacing(4);
         add(lineCat);
         addSpacing(4);
         add(magnitude_Sizer);
      }

      rightBox.margin = 4;
      rightBox.add(this.objName);
      rightBox.addSpacing(4);
      rightBox.add(this.objEdit);
      rightBox.addSpacing(4);
      rightBox.add(this.btnSesame);
      rightBox.addStretch();
      rightBox.add(this.btnClearSolution);

      sizer = new HorizontalSizer();
      sizer.add(leftBox);
      sizer.add(rightBox);
   }

   this.MinMag = new NumericEdit(this);
   with (this.MinMag)
   {
      autoEditWidth = true;
      label.text = "min";
      setPrecision( 4 );
      setRange(3.0, 20.0);
      setReal(true);
      setValue(parms.minMag);
      onValueUpdated = function( value )
      {
         parms.minMag = value;
         cc.magnitudesSet = true;
         cc.joined = false;
      }
   }

   this.MaxMag = new NumericEdit(this);
   with (this.MaxMag)
   {
      autoEditWidth = true;
      label.text = "max";
      setPrecision( 4 );
      setRange(3.0, 20.0);
      setReal(true);
      setValue(parms.maxMag);
      onValueUpdated = function( value )
      {
         parms.maxMag = value;
         cc.magnitudesSet = true;
         cc.joined = false;
      }
   }

   this.numSensitivity = new NumericControl(this);
   with (this.numSensitivity)
   {
      real = true;
      label.text = "Sensitivity:";
      setRange(-3, 3);
      slider.minWidth = 250;
      setPrecision(2);
      setValue( parms.sensitivity );
      toolTip = "<p><b>Star detection sensitivity</p>" +
      "<p>Increase the value to detect less stars.</p>" +
      "<p>Double click to set to default = 0</p>";
	   onValueUpdated = function( value )
      {
         parms.sensitivity = value;
         cc.reScan = true;
      }
      onMouseDoubleClick = function( x, y, buttonState,  modifiers )
		{
         cc.reScan = true;
         parms.sensitivity = 0;
         dlg.numSensitivity.setValue(parms.sensitivity);
      }
   }

   this.maxStarsLabel = new Label(this);
   with (this.maxStarsLabel)
   {
      enabled = !parms.allStars;
      text = "Max. stars:";
      toolTip = "<p><b>Star count limit</p>" +
      "<p>Choose the maximum number of stars for the evaluation process.</p>" +
      "<p>Stars are sorted by their intensity in descending order.</p>"
      onMouseDoubleClick = function( x, y, buttonState,  modifiers )
		{
         parms.maxStars = 500;
         dlg.maxStarsSlider.value = parms.maxStars / 100;
         dlg.maxStarsEdit.text = parms.maxStars.toString();
      }
   }

   this.maxStarsEdit = new Edit(this);
   with (this.maxStarsEdit)
   {
      enabled = !parms.allStars;
      text = parms.maxStars.toString();
      setScaledMaxWidth(dlg.numSensitivity.edit.width);
      onTextUpdated = function( text )
      {
         parms.maxStars = text.toInt();
      }
   }

   this.maxStarsSlider = new Slider(this);
   with (this.maxStarsSlider)
   {
      enabled = !parms.allStars;
      setRange(1, 10) ;
      tickInterval = 1;
      tickStyle = TickStyle_Top ;
      value = parms.maxStars / 100;
	   onValueUpdated = function( value )
      {
         parms.maxStars = value * 100;
         dlg.maxStarsEdit.text = parms.maxStars.toString();
      }
   }

   this.allStars = new CheckBox(this);
   with (this.allStars)
   {
      checked = parms.allStars;
      text = "All";
      toolTip = "<p><b>Star count limit</p>" +
      "<p>If checked, uses all stars for the evaluation process.</p>";
      onCheck = function(checked)
      {
         parms.allStars = checked;
         dlg.maxStarsLabel.enabled = !parms.allStars;
         dlg.maxStarsEdit.enabled = !parms.allStars;
         dlg.maxStarsSlider.enabled = !parms.allStars;
      }
   }


   var maxStarsControls = new HorizontalSizer();
   maxStarsControls.add(this.maxStarsLabel);
   maxStarsControls.addSpacing(4);
   maxStarsControls.add(this.maxStarsEdit);
   maxStarsControls.addSpacing(4);
   maxStarsControls.add(this.maxStarsSlider);
   maxStarsControls.addSpacing(8);
   maxStarsControls.add(this.allStars);

   this.gbxStars = new GroupBox(this);
   with (this.gbxStars)
   {
      title = "Image star limits";
      sizer = new VerticalSizer();
      sizer.margin = 4;
      sizer.add(this.numSensitivity);
      sizer.addSpacing(8);
      sizer.add(maxStarsControls);
   }

   this.numLowerLimit = new NumericEdit(this);
   with (this.numLowerLimit)
   {
      autoEditWidth = true;
      label.text = "Lower limit:";
      setPrecision( 7 );
      setRange(0.0, 1.0);
      setReal(true);
      setValue(parms.lowerLimit);
      onValueUpdated = function( value )
      {
         parms.lowerLimit = value;
      }
   }

   this.numUpperLimit = new NumericEdit(this);
   with (this.numUpperLimit)
   {
      autoEditWidth = true;
      label.text = "Upper limit:";
      setPrecision( 7 );
      setRange(0.0, 1.0);
      setReal(true);
      setValue(parms.upperLimit);
      onValueUpdated = function( value )
      {
         parms.upperLimit = value;
      }
   }


   this.numTarget = new NumericEdit(this);
   with (this.numTarget)
   {
      autoEditWidth = true;
      label.text = "Target:";
      setPrecision( 7 );
      setRange(0.0, 1.0);
      setReal(true);
      setValue(parms.target);
      onValueUpdated = function( value )
      {
         parms.target = value;
      }
   }


   this.gbxBackground = new GroupBox(this);
   with (this.gbxBackground)
   {
      title = "Background";

      sizer = new HorizontalSizer();
      with (sizer)
      {
         margin = 8;
         addStretch();
         add(this.numLowerLimit);
         addSpacing(16);
         add(this.numUpperLimit);
         addSpacing(16);
         add(this.numTarget);
         addStretch();
      }
   }


   this.catLabel = new Label(this);
   this.catLabel.text = "Catalog";
   this.catLabel.textAlignment = TextAlign_Left | TextAlign_VertCenter;

   this.catSel = new ComboBox(this);
   with (this.catSel)
   {
      if (TypeDescription.externalObjects.indexOf("Gaia") > -1) addItem(GAIA3);
      addItem(GAIA);
      addItem("APASS");
      addItem(SDSS);
      for (var i = 0; i < numberOfItems ; i++)
      {
         if (itemText(i) == parms.catalogName)
         {
            currentItem = i;
            break;
         }
      }

      onItemSelected = function( itemIndex )
      {
         cc.catalogValid = false;
         parms.catalogName = itemText(itemIndex);
         Console.writeln('parms.catalogName '+parms.catalogName);
         cc.catalogType = getCatalogType(parms.catalogName);
         Console.writeln('cc.catalogType '+cc.catalogType);
         dlg.cmbSites.enabled = dlg.catSel.itemText(itemIndex) != GAIA3;
         dlg.cbxReload.enabled = dlg.catSel.itemText(itemIndex) != GAIA3;
      }
   }

   this.numStars = new NumericEdit(this);
   with (this.numStars)
   {
      autoEditWidth = true;
      label.text = "Number of stars";
      toolTip = "<p><b>Number of stars / 1^2 °</b></p>" +
         "<p>Set here the number of stars wanted for 1 square degree</p>";
      setRange(50, 10000);
      setReal(false);
      setValue(parms.numStars);
      onValueUpdated = function( value )
      {
         parms.numStars = value;
      }
   }

   this.catalogBox = new GroupBox(this);
   with (this.catalogBox)
   {
      title = "Photometry";

      var hs = new HorizontalSizer();
      hs.margin = 4;
      hs.add(this.catLabel);
      hs.addSpacing(8);
      hs.add(this.catSel);
      hs.addStretch();
      hs.add(this.MinMag);
      hs.addSpacing(12);
      hs.add(this.MaxMag);

      sizer = new VerticalSizer();
      sizer.margin = 8;
      sizer.add(hs);
   }
   // **************************************************************************
   //
   // actions
   //
   // **************************************************************************
   this.btnCalibrate = new PushButton(this);
   with (this.btnCalibrate)
   {
      icon = this.scaledResource(":/icons/execute.png");
      text = "Calibrate";
      onPress = function()
      {
         Console.show();

         dlg.btnCalibrate.enabled = false;

         var eqPos = getImageEQPosition(cc.window);
         cc.ra  = eqPos.x;
         cc.dec = eqPos.y;
         cc.resolution = getImageResolution(cc.window) * 3600.0;
         cc.rotation   = getImageRotation(cc.window);

         Console.writeln("EQPosition " + cc.ra + ", " + cc.dec);
         Console.writeln("Resolution " + cc.resolution);
         Console.writeln("Rotation   " + cc.rotation);

         var rd = cc.window.imageToCelestial(0, 0);
         cc.radius = Math.roundTo(angularSeparation(cc.ra, cc.dec, rd.x, rd.y) * 60, 2);
         cc.plotted = false;
         enabled = false;
         dlg.catSel.enabled = false;
         dlg.numStars.enabled = false;
         dlg.numSensitivity.enabled = false;

         // *****************************************************************
         // equalize channels to a common background
         // *****************************************************************

         backgroundgNeutralization(view, parms.lowerLimit, parms.upperLimit, parms.target);

         if (cc.filterSet)
         {
            cc.joined = false;
            cc.filterSet = false;
         }
         // ********************************************************************
         // check the catalog
         // ********************************************************************
         if (!cc.catalogValid || cc.magnitudesSet)
         {
            cat = createCatalog(parms, cc);
            cc.catalogValid = cat.valid;
            if (!cat.valid)
            {
               errMessage("Catalog creation failed");
               //
               // enable controls
               //
               dlg.btnCalibrate.enabled = true;
               dlg.catSel.enabled = true;
               dlg.numStars.enabled = true;
               dlg.numSensitivity.enabled = true;
               //dlg.gbxM.enabled = true;
               return;
            }
            else
            {
               cc.joined = false;
            }
            cc.magnitudesSet = false;
         }
         // ********************************************************************
         // scan images
         // ********************************************************************
         if (cc.reScan)
         {
            Console.writeln("Scan views with sensitivity " + parms.sensitivity);
            starCollections = scanner(view, parms.sensitivity);
            cc.reScan = false;
            joinedFluxes = joinFluxes(parms, starCollections, cat, cc);
         }
         // ********************************************************************
         // join star tables with catalog, if not done before
         // ********************************************************************
         if (!cc.joined)
            joinedFluxes = joinFluxes(parms, starCollections, cat, cc);

         // ********************************************************************
         // transform catalog magnitudes (SDSS, APASS) or Gaia fluxes
         // ********************************************************************
         transformFluxes(cat);

         // ********************************************************************
         // apply fluxes to star records
         // ********************************************************************
         copyCatalogFluxes(parms, joinedFluxes, cat);
         Console.writeln("Joined star count " + joinedFluxes.length);

         // ********************************************************************
         // sort by catalog Gfluxes and limit the collection
         // ********************************************************************
         if (!parms.allStars & parms.maxStars < joinedFluxes.length)
         {
            joinedFluxes = limitToMaxStars(joinedFluxes, parms.maxStars * cc.circleFactor);
            Console.writeln("Reduced to max. star count " + joinedFluxes.length);
         }

         // ********************************************************************
         // analysis
         // ********************************************************************
         Analysis = new analysis(joinedFluxes);

         // ********************************************************************
         // calibration
         // ********************************************************************

         dlg.btnCalibrate.enabled = true;

         if (!Analysis.valid)
         {
            Console.writeln("Calibration failed");
            dlg.btnCalibrate.enabled = true;
            dlg.catSel.enabled = true;
            dlg.numStars.enabled = true;
            dlg.numSensitivity.enabled = true;
           // dlg.gbxM.enabled = true;
         }
         else
         {
            // *****************************************************************
            // calibrate with PixInsight standard process
            // *****************************************************************
            calibration(view, Analysis, parms);

            // *****************************************************************
            // finally make up the view
            // *****************************************************************
            ApplyAutoSTF(view, DEFAULT_AUTOSTRETCH_SCLIP, DEFAULT_AUTOSTRETCH_TBGND);

            // ********************************************************************
            // plot regression & preview
            // ********************************************************************
            Plotter(view, Analysis, parms, cc);
         }

         // ********************************************************************
         // save parms
         // ********************************************************************
         dlg.btnUndo.enabled = true;
         Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));

         dlg.lblResult.text = "White balance ratios <b>R : G : B = " +
               Analysis.r.toFixed(4) + ' : ' +
               Analysis.g.toFixed(4) + ' : ' +
               Analysis.b.toFixed(4) + "</b>";

         Console.hide();

         cc.success = true;
      }
   }

   this.btnUndo = new PushButton(this);
   with (this.btnUndo)
   {
      enabled = false;
      icon = this.scaledResource(":/icons/undo.png");
      text = "Undo";
      onPress = function()
      {
         dlg.btnCalibrate.enabled = true;
         dlg.catSel.enabled = true;
         dlg.numStars.enabled = true;
         dlg.numSensitivity.enabled = true;
         cc.window.undo();  // undo from colorcalibration
         cc.window.undo();  // undo from backgroundNeutralization
         if (cc.windowRG != null) cc.windowRG.forceClose();
         if (cc.windowRB != null) cc.windowRB.forceClose();
         if (cc.preView  != null) cc.preView.forceClose();
         cc.windowRG = null;
         cc.windowRB = null;
         cc.preView  = null;
         cc.success = false;
         dlg.btnUndo.enabled = false;

         var pos = getImageEQPosition(cc.window);
         Console.writeln('Eqpos ' + pos);
      }
   }

   this.btnOK = new PushButton(this);
   with (this.btnOK)
   {
      icon = this.scaledResource(":/icons/ok.png");
      text = "OK";
      onPress = function()
      {
         if (cc.success)
         {
            // update DATE-OBS
            var obsdaykey = new FITSKeyword("DATE-OBS",
               '\'' + cc.dateobs.toISOString().split('T')[0]  + '\'',
               '[ISO 8601] UTC date/time of exposure start');

            var keywords = [];

            for (var i in cc.window.keywords)
            {
               if (cc.window.keywords[i].name != 'DATE-OBS') keywords.push(cc.window.keywords[i]);
            }

            keywords.push(obsdaykey);
            cc.window.mainView.beginProcess(UndoFlag_Keywords);
            cc.window.keywords = keywords;
            cc.window.mainView.endProcess();
         }

         Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));
         Plotter(view, Analysis, parms, cc);
         dlg.done(0);
      }
   }

   var btn1HSizer = new HorizontalSizer();
   btn1HSizer.margin = 4;
   btn1HSizer.addStretch();
   btn1HSizer.add(this.btnCalibrate);
   btn1HSizer.addStretch();

   var btn2HSizer = new HorizontalSizer();
   btn2HSizer.margin = 4;
   btn2HSizer.add(this.btnUndo);
   btn2HSizer.addStretch();
   btn2HSizer.add(this.btnOK);

   this.lblResult = new Label(this);
   this.lblResult.useRichText = true;
   this.lblResult.textAlignment = TextAlign_HorzCenter | TextAlign_VertCenter;

   this.plotCheck = new CheckBox(this);
   with (this.plotCheck)
   {
      text = "Plot regression lines";
      enabled = cc.gnuplotExe != "";
      if (enabled) checked = parms.plot;

      onCheck = function( checked )
      {
         parms.plot = checked;
      }
   }

   this.createPreview = new  CheckBox(this);
   with (this.createPreview)
   {
      checked = parms.createPreview;
      text = "Create a preview";

      onCheck = function( checked )
      {
         parms.createPreview = checked;
      }
   }

   this.cbxReload = new CheckBox(this);
   with (this.cbxReload)
   {
      enabled = dlg.catSel.itemText(dlg.catSel.currentItem) != GAIA3;
      text = "force catalog re-load";
      //visible = false;
      toolTip = "<p>If checked, the catalog will be reloaded</p>" +
      "<p>Recommended, if the image is different in size from a former solution or after a new solve process</p>";

      onCheck = function( checked )
      {
         cc.reloadCatalog = checked;
      }
   }

   this.lblSites = new Label(this);
   this.lblSites.text = "Vizier server site:";
   this.lblSites.textAlignment = TextAlign_Left | TextAlign_VertCenter;

   this.cmbSites = new ComboBox(this);
   with (this.cmbSites)
   {
      for (var i in sites) addItem(sites[i]);
      if (parms.VizierSite == "") parms.VizierSite = sites[0];
      currentItem = sites.indexOf(parms.VizierSite);
      enabled = dlg.catSel.itemText(dlg.catSel.currentItem) != GAIA3;

      onItemSelected = function( itemIndex )
      {
         parms.VizierSite = sites[itemIndex];
      }
   }


   var options1 = new HorizontalSizer();
   options1.margin = 4;
   options1.add(this.plotCheck);
   options1.addSpacing(8);
   options1.add(this.createPreview);
   options1.addStretch();
   options1.add(this.cbxReload);

   var catalogSizer = new HorizontalSizer();
   catalogSizer.margin = 4;
   catalogSizer.addSpacing(8);
   catalogSizer.add(this.numStars);
   catalogSizer.addSpacing(8);
   catalogSizer.add(this.lblSites);
   catalogSizer.addSpacing(8);
   catalogSizer.add(this.cmbSites);


   this.optionsBox = new GroupBox(this);
   with (this.optionsBox)
   {
      title = "Options";

      sizer = new VerticalSizer();
      sizer.add(options1);
      sizer.addSpacing(8);
      sizer.add(catalogSizer);
   }
   // **************************************************************************
   //
   // assemble GUI
   //
   // **************************************************************************

   this.sizer = new VerticalSizer();
   with (this.sizer)
   {
      margin = 4;
      add(topLineSizer);
      addSpacing(8);
      add(this.lblView);
      addSpacing(8);
      add(this.astrometryBox);
      addSpacing(8);
      add(this.catalogBox);
      addSpacing(8);
      add(this.gbxStars);
      addSpacing(8);
      add(this.gbxBackground);
      addSpacing(8);
      add(btn1HSizer);
      addSpacing(8);
      add(this.lblResult);
      addSpacing(8);
      add(this.optionsBox);
      addSpacing(8);
      add(btn2HSizer);
   }

   this.windowTitle = ID + " " + VERSION;

	this.adjustToContents();
 	this.userResizable = false;
   processEvents();

   // copy key data to edit fields

   var d = getKeyString(cc.window, "DATE-OBS");
   if (d != null)
   {
      cc.dateobs = new Date(d);
   }
   else
   {
      var fi = new FileInfo(cc.window.filePath);
      if (fi.exists) cc.dateobs = fi.lastModified;
      else
         cc.dateobs  = new Date(Date.now());
   }
   this.editDate.text = cc.dateobs.toISOString().split('T')[0];
   cc.epoch = julianDay(cc.dateobs);

   var fLen = getKeyValue(cc.window, "FOCALLEN");
   if (fLen != null) parms.focalLength = fLen;
   this.editFocal.text = parms.focalLength.toFixed(3);

   var pixSz = getKeyValue(cc.window, "XPIXSZ");
   if (pixSz != null) parms.pixelSize = pixSz;
   this.editPixSz.text = parms.pixelSize.toFixed(1);

   cc.objectName = getKeyString(cc.window, "OBJECT");
   if (cc.objectName != null) this.objEdit.text = cc.objectName;

   this.btnSesame.hasFocus = true;
}

function calibration(view, Analysis, parms)
{
   var description = TITLE + ", version " + VERSION + ", catalog " +
                     parms.catalogName;

   colorCalibration(view, Analysis.r, Analysis.g, Analysis.b, parms.upperLimit,
      description);

   Console.noteln("White balance ratios R : G : B = " +
      Analysis.r.toFixed(4) + ' : ' +
      Analysis.g.toFixed(4) + ' : ' +
      Analysis.b.toFixed(4));
}

// *******************************************************************************
// *******************************************************************************
// *******************************************************************************
//
// extract stars with x, y-location and RGB-fluxes by image scanner
//
// *******************************************************************************
// *******************************************************************************
// *******************************************************************************
function analysis(joinedFluxes)
{
   Console.writeln("analyse data");

   this.r = 1;

   this.xListRG = [];
   this.yListRG = [];
   //
   // R : G
   //
   for (var i in joinedFluxes)
   {
      with (joinedFluxes[i])
      {
         this.xListRG.push(iRflux / iGflux); // image
         this.yListRG.push(cRflux / cGflux); // catalog
      }
   }

   //for (var i in  this.xListRG) Console.writeln(i+'\t'+this.xListRG[i]+'\t'+this.yListRG[i]);

   this.regressionRG = regress (this.xListRG, this.yListRG);
   Console.writeln("regressionRG "+'\t'+ this.regressionRG.offset+'\t'+this.regressionRG.gain);

   this.g = (1 - this.regressionRG.offset) / this.regressionRG.gain;

   var maxX = 0;

   for (var i in this.xListRG) if (this.xListRG[i] > maxX) maxX = this.xListRG[i];

   var maxY = this.regressionRG.offset + maxX * this.regressionRG.gain;

   var R = new Rect(0, 0, maxX, maxY);

   this.pointsRG = [];
   for (var i in this.xListRG)
   {
      var p = new Point(this.xListRG[i], this.yListRG[i]);
      if (R.includes(p))
         this.pointsRG.push(new Point(this.xListRG[i], this.yListRG[i]));
   }

   this.xListRB = [];
   this.yListRB = [];
   //
   // R : B
   //
   for (var i in joinedFluxes)
   {
      with (joinedFluxes[i])
      {
         this.xListRB.push(iRflux / iBflux); // image
         this.yListRB.push(cRflux / cBflux); // catalog
      }
   }
   this.regressionRB = regress (this.xListRB, this.yListRB);

   this.b = (1 - this.regressionRB.offset) / this.regressionRB.gain;

   for (var i in this.xListRB) if (this.xListRB[i] > maxX) maxX = this.xListRB[i];

   var maxY = this.regressionRB.offset + maxX * this.regressionRB.gain;

   R = new Rect(0, 0, maxX, maxY);

   this.pointsRB = [];
   for (var i in this.xListRB)
   {
      var p = new Point(this.xListRB[i], this.yListRB[i]);
      if (R.includes(p))
         this.pointsRB.push(new Point(this.xListRB[i], this.yListRB[i]));
   }
   //
   // normalize
   //
   var max = Math.max(this.g, this.b, 1);
   //Console.writeln(this.g+'\t'+ this.b+'\t'+this.r);
   this.r = 1 / max;
   this.g /= max;
   this.b /= max;

   this.valid = this.r <= 1 && this.g <= 1 && this.b <=1;
}

function scanner(view, sensitivity)
{
   //
   // split view, scan the views and return a list of posititions an fluxes
   // per channel (starsLists[collectionR, collectionG, collectionB])
   //
   var detector = new StarDetector();
   detector.sensitivity = Math.pow(10.0, sensitivity);
   detector.upperLimit  = 0.95;
   var starsLists = [0, 0, 0];

   for (var channel = 0; channel < 3; channel++)
   {
      view.image.selectedChannel = channel;
      var img = new Image(view.image);
      var stars = detector.stars(img);
      for (var i in stars)
      {
         stars[i]["catIndex"]    = -1;    // append new properties
         stars[i]["saturated"]   = false; //
      }
      starsLists[channel] = stars;
      processEvents();
   }
   //
   // return 3 lists
   //
   return starsLists;
}

function saturated(star, image)
{
   var xc = Math.floor(star.pos.x);
   var yc = Math.floor(star.pos.y);

   var x0 = xc - 1;
   var y0 = yc - 1;
   var x1 = xc + 1;
   var y1 = yc + 1;

   if (x0 < 0) x0 = 0;
   if (x0 >= image.width) x0 = image.width;

   if (x1 < 0) x1 = 0;
   if (x1 >= image.width) x1 = image.width;

   if (y0 < 0) y0 = 0;
   if (y0 >= image.height) y0 = image.height;

   if (y1 < 0) y1 = 0;
   if (y1 >= image.height) y1 = image.height;

   var r = new Rect(x0, y0, x1, y1);

   var pix = [];

   image.getSamples(pix, r);

   for (var i in pix)
   {
      if (pix > 0.83) return false;
   }
   return true;
}

function joinFluxes(parms, starCollections, cat, cc)
{
   //
   // join RGB channels
   // apply catalog index to identified stars
   //
   // output table format:
   //   catIndex, x, y, iRflux, iGflux, iBflux, cRflux, cGflux, cBflux
   //
   Console.writeln("Join tables");
   for (var channel in starCollections)
   {
      var count = 0;
      // reset flags in catalog
      for (var i in cat.data) cat.data[i].applied = false;
      // pick each star and try finding a catalog record
      var starList = starCollections[channel];
      for (var i in starList)
      {
         var iStar = starList[i];
         var min = 1e+20;
         var k = -1;
         for (var j = 0; j < cat.data.length; j++)
         {
            var cStar = cat.data[j];
            if (!cStar.applied)
            {
               var d = Math.pow((iStar.pos.x - cStar.x), 2) +
                       Math.pow((iStar.pos.y - cStar.y), 2);
               if (d < min)
               {
                  min = d;
                  k = j;
               }
            }
         }
         if (min < 3.99)     // < 2 pixel ^2
         {
            cat.data[k].applied = true;
            iStar.catIndex = k;
            count += 1;
            //Console.writeln(i + '\t' + count + '\t' + Math.sqrt(min));
         }
         else
         {
            iStar.catIndex = -min;
         }
      }
      Console.writeln("Channel# " + channel + " star matches " + count + " from "
         + starCollections[channel].length + " total");
   }
   //
   // make table array
   //
   var joinedFluxes = [];
   var rStars = starCollections[0];
   var gStars = starCollections[1];
   var bStars = starCollections[2];

   for (var i in rStars)
   {
      var s = rStars[i];
      if (s.catIndex < 0) continue;
      for (var j in gStars)
      {
         if (s.catIndex == gStars[j].catIndex)
         {
            for (var k in bStars)
            {
               if (s.catIndex == bStars[k].catIndex)
               {
                  var cRec = cat.data[s.catIndex];    // catalog record
                  var f = {"catIndex":s.catIndex, "x":cRec.x, "y":cRec.y,
                     "iRflux":s.flux, "iGflux":gStars[j].flux, "iBflux":bStars[k].flux,
                     "cRflux":0, "cGflux":0, "cBflux":0};
                  joinedFluxes.push(f);
               }
            }
         }
      }
   }
   cc.joined = true;
   return joinedFluxes;
}

function limitToMaxStars(joinedFluxes, maxStars)
{
   //
   // sort descending by catalog flux "cGflux"
   //
   var sorted = [];
   for (var i in joinedFluxes) sorted.push(joinedFluxes[i]);
   //
   // Sort the numbers in the array in descending order
   //
   sorted.sort(function(a, b){return b.cGflux-a.cGflux});
   while (sorted.length > maxStars) sorted.pop();
   return sorted;
}

function copyCatalogFluxes(parms, joinedFluxes, cat)
{
   for (var i in joinedFluxes)
   {
      var s = joinedFluxes[i];
      var cRec = cat.data[s.catIndex];    // catalog record
      with (cRec)
      {
         s.cRflux = Rflux;
         s.cGflux = Gflux;
         s.cBflux = Bflux;
      }
   }
}

function transformFluxes(cat)
{
   //
   // set cR|G|Bflux with catalog specific transformations
   //
   // ref. http://astro.pas.rochester.edu/~aquillen/ast142/costanti.html
   //
   //	Band 	Eff. Wave. 	mag zpt 	zero flux (Jy) 	zero flux 	Absorption
   //	U 	3600 	-20.94 	1884 	4.19e-9 	1.57
   //	B 	4400 	-20.45 	4646 	6.60e-9 	1.34
   //	V 	5500 	-21.12 	3953 	3.55e-9 	1
   //	R 	6600 	-21.61 	2875 	2.25e-9 	0.75

   const b25_g = -0.06;
   const b25_r = -0.035;

   if (cat.id == "APASS")
   {
      // http://classic.sdss.org/dr7/algorithms/jeg_photometric_eq_dr1.html#usno2SDSS
      // transform u'g'r'i'z -> ugriz (only g and r)
      // g'mag renamed ro gmag
      // r'mag renamed to rmag
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            var g    = g_mag - b25_g * ((g_mag - r_mag) - 0.53);
            var r    = r_mag - b25_r * ((r_mag - i_mag) - 0.21);
            var Rmag = r - 0.1837 * (g - r) - 0.0971;
            Rflux = flux(Rmag, -21.61);
            Gflux = flux(Vmag, -21.12);
            Bflux = flux(Bmag, -20.45);
         }
      }
   }
   else if (cat.id == GAIA)
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            Rflux = flux(RPmag, -21.61);
            Gflux = flux(Gmag,  -21.03);
            Bflux = flux(BPmag, -20.45);
         }
      }
   }
   else if (cat.id == GAIA3)
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            Rflux = flux(RPmag, -21.61);
            Gflux = flux(Gmag,  -21.03);
            Bflux = flux(BPmag, -20.45);
         }
      }
   }
   else if (cat.id == SDSS)
   {
      // https://www.sdss3.org/dr8/algorithms/sdssUBVRITransform.php
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            var Bmag = gmag + 0.313  * (gmag - rmag) + 0.2271;
            var Gmag = gmag - 0.5784 * (gmag - rmag) - 0.0038;
            var Rmag = rmag - 0.1837 * (gmag - rmag) - 0.0971;
            Rflux = flux(Rmag, -21.61);
            Gflux = flux(Gmag, -21.12);
            Bflux = flux(Bmag, -20.45);
         }
      }
   }
   else
   {
      Console.writeln('transformFluxes failed, wrong catalog');
   }
}

function flux(mag, zeropoint)
{
   return Math.pow(10, -(mag - zeropoint) * 0.4);
}
//
// Catalog object and functions
//
// *******************************************************************************
// *******************************************************************************
// *******************************************************************************
//
// catalog object
//
function catalog(filename, catalogName, cc)
{
   //
   // read .tsv files
   //
   this.filename = filename;
   if (!File.exists(filename)) return;
   this.columns = [];
   this.data    = [];
   this.ra      = 0;
   this.dec     = 0;
   this.radius  = 0;
   this.minMag  = 0;    // Gmag
   this.maxMag  = 0;
   this.outMax  = 0;    // max. # of lines
   this.count   = 0;
   this.source  = "";
   this.id      = cc.catalogType.id;
   this.valid   = false;

   var dataFlag = false;

   var ctr = new VOTableReader(filename, cc.catalogType.properties);

   ctr.open();

   if (!ctr.isOpen()) return;

   var header     = ctr.getHeader();
   this.columns   = header.columns;
   this.count     = header.records;
   this.minMag    = header.minMag;    // Gmag
   this.maxMag    = header.maxMag;
   this.source    = header.source;
   this.ra        = header.ra;
   this.dec       = header.dec;
   this.radius    = header.radius;

   while (!ctr.eof())
   {

      while (!ctr.eof())
      {
         var o = ctr.readObject();

         if (o == null) continue;

         var keys = Object.keys(o);

         var skip = false;
         for (var j in keys) // this.columns)
         {
            if (o[keys[j]] == null) //[this.columns[col]] == null)
            {
               skip = true;
               break;
            }
         }
         if (skip) continue;
         //
         // add new properties
         //
         o["Rflux"]  = 0;
         o["Gflux"]  = 0;
         o["Bflux"]  = 0;

         o["applied"] = false;
         o["x"] = 0;
         o["y"] = 0;
         this.data.push(o);
      }
   }

   ctr.close();

   this.valid = this.count > 0;// && catalogName == this.source;
   Console.writeln("CatalogName: " + catalogName + ", " + this.source);
   Console.writeln("Catalog ready, count: " + this.data.length
         + ", min mag: " + this.minMag + ", max mag: " + this.maxMag);

   this.transform = function(cc)
   {
      //
      // remove: if true, records outside image will be removed
      //
      Console.writeln('Transform ' + this.data.length + ' coordinates');
      var temp = [];
      var frame = new Rect(cc.width, cc.height);
      //var raName = cc.columns[0];
      //var deName = cc.columns[1];
      for (var i = 0; i < this.data.length; i++)
      {
         var s = this.data[i];
         var xy = cc.window.celestialToImage(s.ra, s.dec);
         if (frame.includes(xy.x, xy.y))
         {
            s.x = xy.x;
            s.y = xy.y;
            temp.push(s);
         }
      }

      this.data = temp;
      Console.writeln("Used stars: " + this.data.length);
   }
}
// *******************************************************************************
//
// generate image catalog filename from ra & dec
//
// *******************************************************************************
function catalogFilenameCreate(cc, workingDirectory)
{
   var ras = cc.ra.toFixed(6);
   while (ras.length < 10) ras = '0' + ras;
   var des = Math.abs(cc.dec).toFixed(6);
   while (des.length < 9) des = '0' +des;
   if (cc.dec < 0)
      des = '-' + des;
   else
      des = '+' + des;
   var name = workingDirectory + '/' + ras + des + '.' + cc.catalogType.sid;
   Console.writeln("Catalog filename: " + name);
   return name;
}
// *******************************************************************************
//
// Download catalog in segments of max 1 x 1 °
//
//
// *******************************************************************************
function downloadSegmented(cc, parms)
{
   //
   // split the full frame into sub_windows
   // each sub_window must bes <= 1 x 1 degree
   //
   parms.boxes = generateBoxes(cc);
   //
   // define a download file for the box
   //
   for (var i in parms.boxes)
   {
      var box = parms.boxes[i];
      box.filename = boxFile(i, parms.workingDirectory);
      box.loaded   = false
      with (box)
      {
         Console.writeln("row " + row + '\t' + "col " + col + '\t' +
            "ra " + DMSangle.FromAngle(ra / 15).ToString(true) + '\t' +
            "dec " + DMSangle.FromAngle(dec).ToString(true) + '\t' +
            "width " + width + '\t' + "height " + height);
      }
   }

   Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));
   //
   // download files
   //
   var goodFiles = 0;
   for (var i = 0; i < parms.boxes.length; i++)
   {
      parms.boxes[i].loaded = downloadCatalogBox(parms.boxes[i],
            parms.maxMag, parms.minMag, parms, cc);
      if (parms.boxes[i].loaded) goodFiles += 1;
      Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));
      cc.progress = (i + 1) / parms.boxes.length;  // 0..1
      Console.writeln("cc.progress "+cc.progress);
      processEvents();
   }

   if (parms.boxes.length == goodFiles)
   {
      if (aggregateBoxes(parms, cc))
      {
         Console.writeln("File created  " + parms.aggregat);
      }
      else
      {
         Console.warningln("One ore more files not loaded.\n" +
         "This is probably a server or internet problem.\n" +
         "Stop this script and re-start later.\n" );
      }
   }
}
// *******************************************************************************
//
// load catalog extract from web catalog service
//
// *******************************************************************************
function downloadCatalogBox(box, maxMag, minMag, parms, cc)
{
   //
   // http://vizier.u-strasbg.fr/vizier/doc/vizquery.htx
   //
   if (maxMag == minMag) return false;
   if (File.exists(box.filename)) File.remove(box.filename);

   //var strConstraints = "Gmag=" + minMag.toFixed(2) + ".." + maxMag.toFixed(2);

   var url = parms.VizierSite + "viz-bin/votable/?-source=" +
      cc.catalogType.tableName +
      "&-c.ra=" + box.ra.toFixed(6) +
      "&-c.dec=" + box.dec.toFixed(6) +
      "&-c.bd=" + box.width.toString() + 'x' + box.height.toString() +
      "&-out.max=" + parms.numStars +
      "&-out=" + cc.catalogType.queryColumns.join(',');
      url += '&' + cc.catalogType.criteriaMagName + '=' +
         parms.minMag.toFixed(2) + ".." + parms.maxMag.toFixed(2);
   var downloader = new FileDownload( url, box.filename );
   Console.writeln("Download "+ box.filename);
   Console.writeln(url);
   Console.flush();
   var e = false;
   try
   {
      downloader.perform();
   }
   catch (ex)
   {
      Console.writeln(ex);
      e = true;
   }
   if (File.exists(box.filename) && !e)
   {
      Console.writeln("Download completed: " + box.filename);
      return true;
   }
   else
      return false;
}

function aggregateBoxes(parms, cc)
{
   //
   // aggregate all box-files into one, filename = parms.aggregat
   //
   if (parms.boxes == null) return true;
   if (parms.boxes.length == 0) return true;

   Console.writeln("Aggregate VOTABLEs tp: " + parms.aggregat);

   var aggrf = File.systemTempDirectory + "/MaxGenAggregat.txt";
   if (File.exists(aggrf)) File.remove(aggrf);

   var aFile = new File();

   aFile.createForWriting(aggrf);

   var header   = null;

   var filesCount = 0;

   var rdr = null;

   for (var i in parms.boxes)
   {
      var box = parms.boxes[i];

      if (box.filename == null) continue;

      if (box.loaded && File.exists(box.filename))
      {
         var temp = File.systemTempDirectory + "/MaxGenTemp.txt";

         Console.writeln("Read VOTABLE: " + box.filename);

         if (votableToText(box.filename,
            temp,
            cc.catalogType.dataColumns, null,
            cc.catalogType.criteriaMagName))
         {
            //
            // read and add records to aggrf
            //
            rdr = new reader(temp);
            rdr.open();
            if (!rdr.isOpen() || rdr.length == 0)
            {
               rdr.close();
               Console.writeln("Open error or size = 0: " + box.filename);
               continue;
            }
            Console.writeln("File length " + rdr.length);

            var hdrStr     = rdr.readLn();

            if (header == null)
            {
               header = JSON.parse(hdrStr);
               //
               // overwrite subimage ra, dec with image center
               //
               header.ra      = cc.ra,
               header.dec     = cc.dec;
               header.radius  = cc.radius;
            }
            else
            {
               var h = JSON.parse(hdrStr);
               header.records += h.records;
            }
            //
            // copy records to aFile
            //
            var cnt = 0;
            while (!rdr.eof())
            {
               var s = rdr.readLn();
               aFile.outTextLn(s);
               cnt += 1;
            }
            rdr.close();
            filesCount += 1;
            Console.writeln("File copied: " + cnt + " records");
            File.remove(temp);
            Console.writeln("File deleted: " + temp);
            Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));
         }
      }
   }
   aFile.close();

   if (header == null) return false;
   //
   // copy updated header and aFile to finale aggregatFile
   //
   if (parms.aggregat == '') return true;

   var aggregatFile = new File();
   aggregatFile.createForWriting(parms.aggregat);

   Console.writeln("Aggreate file " + parms.aggregat);

   var strHeader = JSON.stringify(header);

   Console.writeln("File header   " + strHeader);

   rdr = new reader(aggrf);
   rdr.open();

   var aggregatedRecords = 0;

   if (rdr.length > 0)
   {
      Console.writeln("Read temp aggregat: " + rdr.length + " bytes");
      //
      // write the header
      //
      aggregatFile.outTextLn(strHeader);
      //
      // copy records to aFile
      //
      while (!rdr.eof())
      {
         var s = rdr.readLn();
         if (s.length > 0)
         {
            aggregatFile.outTextLn(s);
            aggregatedRecords += 1;
         }
      }
      Console.writeln("File copied: " + cnt + " records");
   }
   else
   {
      Console.writeln("Temp aggregat is empty: " + aggrf);
   }
   rdr.close();
   aggregatFile.close();

   var valid = parms.boxes.length == filesCount && aggregatedRecords > 10;
   //
   // cleanup
   //
   if (!valid)
   {
      Console.writeln("Delete " + parms.aggregat);
      File.remove(parms.aggregat);
   }

   if (filesCount == parms.boxes.length && parms.boxes.length > 0)
   {
      for (var i in parms.boxes)
      {
         var box = parms.boxes[i];
         Console.writeln("Delete " + box.filename);
         if (File.exists(box.filename)) File.remove(box.filename);
         Settings.write(ID + "/parms", DataType_String, JSON.stringify(parms));
      }
      File.remove(aggrf);
      parms.boxes = [];
      parms.aggregat = '';
      Console.writeln("Subframes aggregated to: " + parms.aggregat);
   }

   return valid;
}

// *******************************************************************************
//
// create catalog from existing file, reload, if not conditioned
//
// *******************************************************************************
function createCatalog(parms, cc)
{

   Console.writeln("CreateCatalog");
   Console.flush();

   var cat = null;      // will become a catalog object

   if (parms.catalogName == GAIA3)
   {
      var folder = File.systemTempDirectory;

      //
      // calc circleFactor
      //
      var areaImage = cc.height / cc.width;
      var areaCircle = Math.sqrt(1 + areaImage * areaImage);
      cc.circleFactor = areaCircle / areaImage;
      var catalogFilename = FluXXLocalDR3(cc.ra, cc.dec, cc.radius / 60.0,
         parms.minMag, parms.maxMag, parms.numStars * cc.circleFactor, folder);
      cat = new catalog(catalogFilename, parms.catalogName, cc);
      if (!cat.valid)
      {
         errMessage("Download error, choose another catalog");
         return cat;
      }
   }
   else
   {
      cc.circleFactor = 1;
      var catalogFilename = fileNearPosition(cc.ra, cc.dec,
                              parms.workingDirectory, cc.catalogType.sid);
      //
      // create the catalog filename
      //
      if (catalogFilename == '')
         catalogFilename = catalogFilenameCreate(cc, parms.workingDirectory);
      //
      // *******************************************************************************
      //
      var load = false;
      if (File.exists(catalogFilename) && !cc.reloadCatalog)
      {
         //
         // read existing
         //
         Console.writeln("Read from file " + catalogFilename);
         cat = new catalog(catalogFilename, parms.catalogName, cc);
      }
      else
         load = true;

      if (load)
      {
         var strFields = "";
         var strConstraints = "";
         Console.writeln("Download new file " + catalogFilename);
         Console.flush();
         //
         //
         //
         parms.aggregat = catalogFilename;

         downloadSegmented(cc, parms)
         //
         // read new file
         //
         cat = new catalog(catalogFilename, parms.catalogName, cc);
         if (!cat.valid)
         {
            errMessage("Download error, choose another catalog");
            return cat;
         }
      }
   }

   if (cat.valid)
   {
      //
      // record selection
      //
      verifyCat(cat, parms)
      cat.transform(cc);
   }
   return cat;
}
// *****************************************************************************
//
// Subframe
//
// *****************************************************************************

function boxFile(index, workingDirectory)
{
   var name = workingDirectory + "/MaskGenBox_" + index.toString() + ".vot";
   if (File.exists(name)) File.remove(name);
   return name;
}

function generateBoxes(cc)
{
   // return a collection of boxes for download

   var view   = cc.window.mainView;
   var image  = view.image;

   var iScale = cc.resolution;

   var pixelD = 1 /( iScale / 3600 );            // pixel per degree

   Console.writeln("iScale  "+iScale );
   Console.writeln("Pixel / Grad " + pixelD);

   var width  = image.width;
   var height = image.height;

   var cp = new Point(width / 2, height / 2);

   var radius = Math.sqrt(Math.pow(cp.x, 2) + Math.pow(cp.y, 2));
   var F = radius * radius * Math.PI;
   Console.writeln("Cone area pix^2 "+F);


   var p1 = new Point(0, 0);
   var p2 = new Point(width, 0);
   var p3 = new Point(0, height);
   var p4 = new Point(width, height);
   var pA = [p1, p2, p3, p4];
   var c1 = cc.window.imageToCelestial(p1);
   var c2 = cc.window.imageToCelestial(p2);
   var c3 = cc.window.imageToCelestial(p3);
   var c4 = cc.window.imageToCelestial(p4);

   Console.writeln(c1);
   Console.writeln(c2);
   Console.writeln(c3);
   Console.writeln(c4);

   var maxDec = Math.max(c1.y, c2.y, c3.y, c4.y);
   var minDec = Math.min(c1.y, c2.y, c3.y, c4.y);
   //
   // rotate by CROTA1
   //
   var angle = cc.rotation;
   Console.writeln("Rotate by "+angle);

   for (var i in pA) pA[i] = rot(pA[i], cp, angle);

   // get min and max x, y

   var minX = Number.MAX_VALUE;
   var minY = Number.MAX_VALUE;
   var maxX = Number.MIN_VALUE;
   var maxY = Number.MIN_VALUE;

   for (var i in pA)
   {
      if (pA[i].x < minX) minX = pA[i].x;
      if (pA[i].y < minX) minY = pA[i].y;
      if (pA[i].x > maxX) maxX = pA[i].x;
      if (pA[i].y > maxY) maxY = pA[i].y;
   }

   var frame = new Rect(0, 0, maxX - minX, maxY - minY);
   var rows = Math.ceil(frame.height / pixelD);
   Console.writeln("Frame around rotated image rectangle: " + frame);
   Console.writeln("Rect area pix^2 " + (frame.width * frame.height));
   Console.writeln("Frame rows of equal declination span: " + rows);

   var crval1 = cc.ra;
   var crval2 = cc.dec;

   var decSpan   = maxDec - minDec; // frame.height * iScale;

   var decTop    = crval2  + decSpan / 2;
   var decBottom = crval2  - decSpan / 2;

   var dDec = (decTop - decBottom) / rows;
   Console.writeln("Declination span: " + dDec);
   Console.flush();

   var boxes = [];

   for (var row = 0; row < rows; row++)
   {
      var rowDecTop    = decTop - row * dDec;
      var rowDecCenter = rowDecTop - dDec / 2;
      var rowDecBottom = rowDecTop - dDec;

      // function raSpan returns: a = high ra, b = low ra, angle a-b

      var topCircle    = raSpan(crval1, rowDecTop,    frame.width, iScale);
      var centerCircle = raSpan(crval1, rowDecCenter, frame.width, iScale);
      var bottomCircle = raSpan(crval1, rowDecBottom, frame.width, iScale);

      var maxCircle = Math.max(arcDiff(topCircle.a, topCircle.b));

      var ra = 0;

      if (topCircle.angle > bottomCircle.angle)
      {
         ra = topCircle.a;    // start of row in this decl band
      }
      else
      {
         ra = bottomCircle.a;    // start of row in this decl band
      }

      if (Math.sign(rowDecTop) != Math.sign(rowDecBottom))
      {
         var equ = raSpan(crval1, 0, frame.width, iScale);
         ra = equ.a;
         maxCircle = Math.max(arcDiff(equ.a, equ.b));
      }

      if (maxCircle > 90) maxCircle = 180 - maxCircle;

      var cols = Math.ceil(maxCircle);       // <= 1° column width

      var boxWidth = maxCircle / cols;

      for (var col = 0; col < cols; col++)
      {
         boxes.push({ra:reduce(ra - boxWidth / 2),
                     dec:rowDecCenter,
                     width:boxWidth,
                     height:rowDecTop-rowDecBottom,
                     row:row,
                     col:col});
         ra -= boxWidth;
         ra = reduce(ra);
      }
   }
   return boxes;
}

function raSpan(ra, dec, pixels, scale)
{
   // get 2 ra positions, left and right from ra
   var cos = Math.cos(dec * Math.RAD);
   var a = ra + (pixels / 2 * scale/3600) / cos;
   var b = ra - (pixels / 2 * scale/3600) / cos;
   a = reduce(a);
   b = reduce(b);
   return {a:a,b:b,angle:arcDiff(a, b)};
}

function arcDiff(a, b)
{
   var z = a - b;
   if (z >= 180 && z < 180)
      return z;
   else
      z = z - Math.floor((z + 180) / 360) * 360;
   return z;
}

function reduce(angle)
{
   while (angle >= 360) {angle -= 360;}
   while (angle <    0) {angle += 360;}
   return angle;
}

function rot(p, center, angle)
{
   angle *= Math.RAD;
   var x = Math.cos(angle) * (p.x - center.x) -
           Math.sin(angle) * (p.y - center.y) + center.x;

   var y = Math.sin(angle) * (p.x - center.x) +
           Math.cos(angle) * (p.y - center.y) + center.y;
   return new Point(x,y);
}

// *****************************************************************************
// *****************************************************************************
// *****************************************************************************


function verifyCat(cat, parms)
{
   var temp = [];    // collect valid records

   if (parms.catalogName == 'APASS')
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {

            if (Vmag < parms.minMag || Vmag > parms.maxMag) continue;
            if (nobs < parms.numObs) continue;
            temp.push(cat.data[i]);
        }
      }
   }
   else if (parms.catalogName == GAIA)
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            if (Gmag < parms.minMag || Gmag > parms.maxMag) continue;
            temp.push(cat.data[i]);
         }
      }
   }
   else if (parms.catalogName == GAIA3)
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            if (Gmag < parms.minMag || Gmag > parms.maxMag) continue;
            temp.push(cat.data[i]);
         }
      }
   }
   else if (parms.catalogName == SDSS)
   {
      for (var i in cat.data)
      {
         with (cat.data[i])
         {
            if (gmag < parms.minMag || gmag > parms.maxMag) continue;
            if (!(cl == 3 || cl == 6)) continue;
            temp.push(cat.data[i]);
         }
      }
   }
   cat.data = temp;
}
// *****************************************************************************
//
// Math functions
//
// *****************************************************************************
function regress (x, y)
{
   //
	// regression Y over X
	//	X = image
	//	Y = catalog
	//
   var offset     = 0;
	var gain       = 0;
   var count      = 0;
	var iter       = 1;
	var regError   = 0;

	var lastC = 1.0;

   while (iter < 10)
   {
      count    = x.length;
      var bc   = linearRegression(x, y);
      gain     = bc.a;
      offset   = bc.b;
      //
      // calc deviation
      //
      var sum = 0;
      for (var i in x) sum += Math.pow(y[i] - (offset + gain * x[i]), 2);
      var stdDev = Math.sqrt(sum / (count - 1));
      var tempX = [];
      var tempY = [];

      for (var i in x)
      {
         if (Math.abs(y[i] - (offset + gain * x[i])) < stdDev)
         {
            tempX.push(x[i]);
            tempY.push(y[i]);
         }
      }
      if (tempX.length == 0) break;
      x = tempX;
      y = tempY;

      if (tempX.length == count) break;

      regError = Math.abs(gain - lastC) / lastC;

      if (regError < 0.01) break;

      lastC = gain;
      iter += 1;
   }

   return {gain:gain, offset:offset, iter:iter, regError:regError,
           xList:tempX, yList:tempY};
}
// *****************************************************************************
//
// FUNCTION linearRegression
//
// *****************************************************************************
function linearRegression(independent, dependent)
{
    let lr = {};

    let independent_mean = arithmeticMean(independent);
    let dependent_mean = arithmeticMean(dependent);
    let products_mean = meanOfProducts(independent, dependent);
    let independent_variance = variance(independent);

    lr.a = (products_mean - (independent_mean * dependent_mean) ) / independent_variance;

    lr.b = dependent_mean - (lr.a * independent_mean);

    return lr;
}

// *****************************************************************************
// FUNCTION arithmeticMean
// *****************************************************************************
function arithmeticMean(data)
{
    let total = 0;

    // note that incrementing total is done within the for loop
    for(let i = 0, l = data.length; i < l; total += data[i], i++);

    return total / data.length;
}

// *****************************************************************************
// FUNCTION meanOfProducts
// *****************************************************************************
function meanOfProducts(data1, data2)
{
    let total = 0;

    // note that incrementing total is done within the for loop
    for(let i = 0, l = data1.length; i < l; total += (data1[i] * data2[i]), i++);

    return total / data1.length;
}

// *****************************************************************************
// FUNCTION variance
// *****************************************************************************
function variance(data)
{
    let squares = [];

    for(let i = 0, l = data.length; i < l; i++)
    {
        squares[i] = Math.pow(data[i], 2);
    }

    let mean_of_squares = arithmeticMean(squares);
    let mean = arithmeticMean(data);
    let square_of_mean = Math.pow(mean, 2);
    let variance = mean_of_squares - square_of_mean;

    return variance;
}
// *****************************************************************************
// End FUNCTION linearRegression
// *****************************************************************************
//
// *****************************************************************************
// STF Auto Stretch routine
// *****************************************************************************
function ApplyAutoSTF( view, shadowsClipping, targetBackground )
{
   var stf = new ScreenTransferFunction;

   var n = 3;

   var median = view.computeOrFetchProperty( "Median" );

   var mad = view.computeOrFetchProperty( "MAD" );
   if (mad == null) throw "ApplyAutoSTF failed, mad is null";
   mad.mul( 1.4826 ); // coherent with a normal distribution

   var c0 = 0, m = 0;
   for ( var c = 0; c < n; ++c )
   {
      if ( 1 + mad.at( c ) != 1 )
         c0 += median.at( c ) + shadowsClipping * mad.at( c );
      m  += median.at( c );
   }
   c0 = Math.range( c0/n, 0.0, 1.0 );
   m = Math.mtf( targetBackground, m/n - c0 );

   stf.STF = [ // c0, c1, m, r0, r1
               [c0, 1, m, 0, 1],
               [c0, 1, m, 0, 1],
               [c0, 1, m, 0, 1],
               [0, 1, 0.5, 0, 1] ];

   stf.executeOn( view, false );
}

function stringOf(chr, len)
{
   var a = [];
   for (var i = 0; i < len; i++) a.push(chr);
   return a.join('');
}

// *******************************************************************************
//
// Calibration processes BackgroundNeutralization and ColorCalibration
//
// *******************************************************************************
function backgroundgNeutralization(view, lowerLimit, upperLimit, target)
{
   var P = new BackgroundNeutralization;
   P.backgroundReferenceViewId = "";
   P.backgroundLow = lowerLimit;
   P.backgroundHigh = upperLimit;
   P.useROI = false;
   P.roiX0 = 0;
   P.roiY0 = 0;
   P.roiX1 = 0;
   P.roiY1 = 0;
   P.mode = BackgroundNeutralization.prototype.TargetBackground;
   P.targetBackground = target;
   P.executeOn(view);
   Console.writeln("backgroundgNeutralization");
   Console.flush();
}

function colorCalibration(view, r, g, b, upperLimit, description)
{
   var P = new ColorCalibration;
   P.whiteReferenceViewId = "";
   P.whiteLow = 0.0000000;
   P.whiteHigh = 0.9000000;
   P.whiteUseROI = false;
   P.whiteROIX0 = 0;
   P.whiteROIY0 = 0;
   P.whiteROIX1 = 0;
   P.whiteROIY1 = 0;
   P.structureDetection = false;
   P.structureLayers = 5;
   P.noiseLayers = 1;
   P.manualWhiteBalance = true;
   P.manualRedFactor = r;
   P.manualGreenFactor = g;
   P.manualBlueFactor = b;
   P.backgroundReferenceViewId = "";
   P.backgroundLow = 0.0000000;
   P.backgroundHigh = upperLimit;
   P.backgroundUseROI = false;
   P.backgroundROIX0 = 0;
   P.backgroundROIY0 = 0;
   P.backgroundROIX1 = 0;
   P.backgroundROIY1 = 0;
   P.outputWhiteReferenceMask = false;
   P.outputBackgroundReferenceMask = false;
   P.setDescription(description);
   P.executeOn(view);
}

// *******************************************************************************
//
// GNUPlot graphical output
//
// *******************************************************************************
function Plotter(view, Analysis, parms, cc)
{
   if (Analysis == null) return;
   if (!cc.plotted)
   {
      if (parms.plot)
      {
         with (Analysis)
         {
            cc.windowRG = plot(pointsRG, regressionRG.gain,
               regressionRG.offset, "#008F00", cc.objectName, parms.catalogName, "G",
               cc.gnuplotExe);
            cc.windowRB = plot(pointsRB, regressionRB.gain,
               regressionRB.offset, "#0000FF", cc.objectName, parms.catalogName, "B",
               cc.gnuplotExe);
         }
      }
      // *****************************************************************
      // create a previev with standard curves
      // *****************************************************************
      if (parms.createPreview) cc.preView = createPreview(view);
   }
   cc.plotted = true;   // prevent re-entry
}

function plot(points, a, b, lineColor, Id, catalog, chrChannel, GNUPlot)
{
   var scriptPath = tempFile("gbnuScript", "plt");
	var imagePath  = tempFile("image", "svg");
	var outputPath = tempFile("output", "txt");
	var dataPath   = tempFile("XYdata", "csv");
	var errorPath  = tempFile("gnuplot_error", "txt");

   var strRatio = "R : " + chrChannel + " = 1 : " + ((1 - b) / a).toFixed(4);

	var title =" set title font \"Times Bold, 20\" enhanced \"FluXX Photometric Calibration";
	title += "\\n{/*0.8 Object " + Id + '}';
	title += "\\n{/*0.8 Catalog " + catalog + " - Color ratio " + strRatio + '}';

   var lines = [];
   for (var u in points) lines.push(points[u].x+'\t'+points[u].y);
   File.writeTextFile(dataPath, lines.join('\n') );

	// write GNUPlot script

	var script = [];
	script.push("verbose");
	script.push("set terminal svg size 600,500 enhanced background rgb 'white' font \"Helvetica,10\"");
	script.push("set output \"" + imagePath + "\"");
	script.push(title);
	script.push("set xlabel \"Image R : " + chrChannel + "\" font \"Helvetica Bold, 14\"");
	script.push("set ylabel \"Catalog R : " + chrChannel + "\" font \"Helvetica Bold, 14\"");
	script.push("set style line 1 lc rgb 'black' lt 1 lw 1 dashtype 0");
	script.push("set style line 2 lc rgb 'black' lt 1 lw 1 dashtype 1");
	script.push("set style line 3 lc rgb 'grey' lt 0 lw 1 dashtype '-'");
	script.push("set grid back ls 3");

	script.push("a = " + a);
   script.push("b = " + b);
   script.push("g(x) = a * x + b");

	script.push("set key left top");

	script.push("plot \"" + dataPath + "\"  using 1:2 title 'Catalog to Image Ratio' with points ls 1, \\");
   script.push("g(x)  lt rgb \"" + lineColor + "\" lw 3");
	script.push("quit\n");

	File.writeTextFile( scriptPath, script.join('\n'));

	var P = new ExternalProcess();
	with (P)
	{
      workingDirectory = File.systemTempDirectory;

		redirectStandardInput ( scriptPath );
		redirectStandardOutput( outputPath );
      redirectStandardError ( errorPath );

		onStarted = function () {Console.writeln("Start " + GNUPlot);}

		onError = function ( errorCode ) {Console.writeln("Error# " + errorCode);}

		onFinished = function ( exitCode, exitStatus )
		{Console.writeln("End Exit code " + exitCode + '\t' + ", Status " + exitStatus);}

      onStandardErrorDataAvailable = function ()
      {
         Console.writeln("onStandardErrorDataAvailable");
      }

      onStandardOutputDataAvailable = function ()
      {
         Console.writeln("onStandardOutputDataAvailable");
      }

      onStateChanged = function ( newState )
      {
         //Console.writeln("onStateChanged " + newState);
      }

     Console.writeln("GNUPlot waitForStarted: " + GNUPlot);

      start(GNUPlot);
      Console.writeln("GNUPlot pid: " + pid);
      waitForStarted();
      waitForFinished();
      terminate();
	}

	// load GNUPlot image

   var imgWin = null;

	if (File.exists(imagePath))
	{
		try
		{
         Console.writeln("Load plot image: " + imagePath);
			var bm = new Bitmap(imagePath);
			imgWin = new ImageWindow(bm.width, bm.height, 3, 32, true, true, ID +"_Plot");
			imgWin.mainView.beginProcess(UndoFlag_NoSwapFile);
			imgWin.mainView.image.blend( bm );
			imgWin.mainView.endProcess();
			imgWin.show();
		}
		catch (ex)
		{Console.writeln("Image load error: " + ex);}
		File.remove( imagePath );
	}
      else Console.writeln("Graphics file " + imagePath + " not created");

	if (File.exists(scriptPath))  File.remove( scriptPath );
	if (File.exists(outputPath))  File.remove( outputPath );
	if (File.exists(dataPath))    File.remove( dataPath );
	if (File.exists(errorPath))   File.remove( errorPath );

   return imgWin;
}

function createPreview(view)
{
   //
   // create a small preview with standard curves
   //
   var img = new Image(view.image);

   var imgWin = new ImageWindow(img.width, img.height, 3, 32, true, true, "FluXX_View");
   imgWin.mainView.beginProcess(UndoFlag_NoSwapFile);
   imgWin.mainView.image.apply( img );
   imgWin.mainView.endProcess();

   var preView = imgWin.mainView;

   ApplyAutoSTF(preView, DEFAULT_AUTOSTRETCH_SCLIP, DEFAULT_AUTOSTRETCH_TBGND );

   var stf = preView.stf;
	var c0  = (stf[0][1] + stf[1][1] + stf[2][1]) / 3;
	var m   = (stf[0][0] + stf[1][0] + stf[2][0]) / 3;
   var h   = (stf[0][2] + stf[1][2] + stf[2][2]) / 3;
//Console.writeln("c0 " + c0);

	var HT = new HistogramTransformation;
 	HT.H  = [[  0, 0.5, 1.0, 0, 1.0],
            [  0, 0.5, 1.0, 0, 1.0],
            [  0, 0.5, 1.0, 0, 1.0],
            [ c0,   m,   h, 0, 1.0],
            [  0, 0.5, 1.0, 0, 1.0]];

   HT.executeOn(preView, false);

   var P = new CurvesTransformation;
   P.R = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.Rt = CurvesTransformation.prototype.AkimaSubsplines;
   P.G = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.Gt = CurvesTransformation.prototype.AkimaSubsplines;
   P.B = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.Bt = CurvesTransformation.prototype.AkimaSubsplines;
   P.K = [ // x, y
      [0.00000, 0.00000], [0.24685, 0.14251],
      [0.75315, 0.87961], [1.00000, 1.00000]
   ];
   P.Kt = CurvesTransformation.prototype.AkimaSubsplines;
   P.A = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.At = CurvesTransformation.prototype.AkimaSubsplines;
   P.L = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.Lt = CurvesTransformation.prototype.AkimaSubsplines;
   P.a = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.at = CurvesTransformation.prototype.AkimaSubsplines;
   P.b = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.bt = CurvesTransformation.prototype.AkimaSubsplines;
   P.c = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.ct = CurvesTransformation.prototype.AkimaSubsplines;
   P.H = [ // x, y
      [0.00000, 0.00000], [1.00000, 1.00000]
   ];
   P.Ht = CurvesTransformation.prototype.AkimaSubsplines;
   P.S = [ // x, y
      [0.00000, 0.00000], [0.49622, 0.86486], [1.00000, 1.00000]
   ];
   P.St = CurvesTransformation.prototype.AkimaSubsplines;

   P.executeOn(preView, false);

   P = new ScreenTransferFunction;
   P.STF = [ // c0, c1, m, r0, r1
   [c0, 1.00000, 0.5, 0.00000, 1.00000],
   [c0, 1.00000, 0.5, 0.00000, 1.00000],
   [c0, 1.00000, 0.5, 0.00000, 1.00000],
   [0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
   ];
   P.interaction = ScreenTransferFunction.prototype.SeparateChannels;

   P.executeOn(preView, false);

   imgWin.show();

   imgWin.zoomFactor = - 6;
   imgWin.fitWindow();
   return imgWin;
}



// *******************************************************************************
//
// Auxiliary functions
//
// *******************************************************************************

function angularSeparation(ra1, de1, ra2, de2)
{
   // degrees to radiant
   var a1 = ra1 * Math.RAD;
   var a2 = ra2 * Math.RAD;
   var d1 = de1 * Math.RAD;
   var d2 = de2 * Math.RAD;
   var z  = Math.acos(Math.sin(d1) * Math.sin(d2) + Math.cos(d1) * Math.cos(d2) * Math.cos(a1 - a2));
   var d  = Math.abs(z) * Math.DEG;
   return d % 360.0;
}

function julianDay(d)
{
   d = new Date(d);
   var y  = d.getFullYear();
   var m  = d.getMonth();
   var D  = d.getDay();
   var jd = 367 * y - (7 * (y + 5001 + (m - 9)/7))/4 + (275 * m)/9 + D + 1729777 ;// Math.complexTimeToJD(y, m , D);
   return jd;
}

function errMessage(txt)
{
   new MessageBox(txt, "FluXX", StdIcon_Error).execute();
}

function imageResolution(focal, pix)
{
   return 206.265 * pix / focal;
}

function getKeyValue(window, keyName)
{
      //
   // search key keyName
   //
   for (var i in window.keywords)
   {
      with (window.keywords[i])
      {
         if (name == keyName)
         {
            return value.toNumber();
         }
      }
   }
   return null;
}

function getKeyString(window, keyName)
{
   //
   // search key keyName
   //
   for (var i in window.keywords)
   {
      with (window.keywords[i])
      {
         if (name == keyName)
         {
            var s = strippedValue;
            return s;
         }
      }
   }
   return null;
}

function getImageEQPosition(window)
{
   if (window.astrometricSolution().length > 0)
   {
      //var ast = window.astrometricSolution();// new astrometricSolution(window.astrometricSolution()[0]);

      var CRVAL1 = getKeyValue(window, "CRVAL1");
      var CRVAL2 = getKeyValue(window, "CRVAL2");
      if (getKeyString(window,"CTYPE1") == "RA---TAN") return new Point(CRVAL1, CRVAL2);
      else return new Point(CRVAL2, CRVAL1);
   }
   else
      return new Point(-2, -1);
}

function getImageResolution(window)
{
   if (window.astrometricSolution().length > 0)
   {
      //var ast = window.astrometricSolution();//var ast = new astrometricSolution(window.astrometricSolution()[0]);
      var CDELT1 = getKeyValue(window, "CDELT1");
      var CDELT2 = getKeyValue(window, "CDELT2");
      return (Math.abs(CDELT1) + Math.abs(CDELT2)) / 2;
   }
   else
      return 0;
}

function getImageRotation(window)
{
   if (window.astrometricSolution().length > 0)
   {
      //var ast = window.astrometricSolution();//var ast = new astrometricSolution(window.astrometricSolution()[0]);
      return getKeyValue(window, "CROTA1");
   }
   else
      return 0;
}

function gnuplotExecutable()
{
   var gnu = File.systemTempDirectory + "/gnuplot.exe";
   if (File.exists(gnu)) return gnu;
   var f = File.currentWorkingDirectory;
   if (!f.endsWith("/bin")) f += "/bin";
   f += "/gnuplot.exe";
   if (File.exists(f))
   {
      File.copyFile( gnu, f );
      return gnu;
   }
   return "";  // not found
}

function fileExists(filename)
{
   if (filename == '') return false;
   return File.exists(filename);
}

function fileNearPosition(ra, dec, workingDirectory, suffix)
{
   var files = getFiles(workingDirectory, suffix);
   //
   // find file with min distance to image center
   //
   var bestFile = '';
   var minAng = 99999;
   for (var i in files)
   {
      var filename = files[i];

      var fInfo = new FileInfo(filename);
      var rd = fInfo.nameAndExtension.substr(0, fInfo.nameAndExtension.length -
            fInfo.suffix.length);
      // split rd in ra & dec
      var ang = [];
      var signIndex = rd.indexOf('+');
      if (signIndex > -1)
      {
         ang.push(rd.substring(0, signIndex));
         ang.push(rd.substring(signIndex));
      }
      else
      {
         signIndex = rd.indexOf('-');
         if (signIndex < 0) continue;
         ang.push(rd.substring(0, signIndex));
         ang.push(rd.substring(signIndex));
      }
      if (ang.length == 2)
      {
         var ra_  = ang[0].toNumber();
         var dec_ = ang[1].toNumber();

         var a = angularSeparation(ra, dec, ra_, dec_);

         if (a < minAng && a < 5 / 3600)
         {
            bestFile = filename;
            minAng = a;
         }
      }
   }
   return bestFile;
}

function getFiles(workingDirectory, suffix)
{
   var names = [];
   var finder = new FileFind();
   finder.begin(workingDirectory + '/' + "*" + suffix);
   while (finder.name && finder.isFile)
   {
      var filename =  workingDirectory + '/' + finder.name;
      names.push(filename);
      finder.next();
   }
   finder.end();
   return names;
}

function tempFile(name, ext)
{
   var workingDir = File.systemTempDirectory;
   workingDir = workingDir.replace('\\', '/');
   if (!workingDir.endsWith('/')) workingDir += '/';
   var filename = "";
   var index = 0;
   while (true)
   {
      if (index == 0)
         filename = workingDir + name + '.' + ext;
      else
         filename = workingDir + name + index.toString() + '.' + ext;
      if (!File.exists(filename)) break;
      index += 1;
   }
   return filename;
}


// *****************************************************************************
//
// VOTABLE convert and read
//
// *****************************************************************************

function VOTableReader(filename, properties)
{
   this.filename = filename;

   var hdr = null;
   var index = -1;

/*
   var rdr = new reader(filename);
   rdr.open();
   if (rdr.isOpen())
   {
      var line = rdr.readLn();
      hdr = JSON.parse(line);
      for (var i in hdr.columns) hdr.columns[i] = properties[i];
   }
   rdr.close();

   this.readObject = function()
   {
      if (!rdr.isOpen()) return null;
      if (rdr.eof())
      {
         rdr.close();
         return null;
      }
      var line = rdr.readLn();
      if (line == '' ) return null;
      //
      // split line and return an object
      //
      var items = line.split('\t');
      var obj = new Object();
      for (var i in hdr.columns)
      {
         if (isNumeric(items[i]))
            obj[hdr.columns[i]] = items[i].toNumber();
         else
            obj[hdr.columns[i]] = items[i];
      }
      return obj;
   }

   this.isOpen = function()
   {
      return rdr.isOpen();
   }

   this.eof = function()
   {
      return rdr.eof();
   }

   this.open = function()
   {
      rdr = new reader(this.filename);
      rdr.open();
      if (rdr.isOpen()) rdr.readLn();  // skip the header
   }

   this.close = function()
   {
      return rdr.close();
   }
*/
   var lines;

   this.readObject = function()
   {
      if (hdr == null) return null;
      index += 1;
      if (index >= lines.length) return null;
      var line = lines[index];
      //
      // split line and return an object
      //
      var items = line.split('\t');
      var obj = new Object();
      for (var i in hdr.columns)
      {
         if (isNumeric(items[i]))
            obj[hdr.columns[i]] = items[i].toNumber();
         else
            obj[hdr.columns[i]] = items[i];
      }
      return obj;
   }

   this.isOpen = function()
   {
      return index > -1;
   }

   this.eof = function()
   {
      return index >= lines.length;
   }

   this.open = function()
   {
      lines = File.readLines(filename);

      if (lines.length > 0)
      {
         index = 0;
         hdr = JSON.parse(lines[index]);
         for (var i in hdr.columns) hdr.columns[i] = properties[i];
      }
   }

   this.close = function()
   {
      lines = null;
      index = -1;
      return true;
   }

   this.getHeader = function()
   {
      return hdr;
   }
}

function  votableToText(fIn, fOut, catColumnsIn, catColumnsOut, magName)
{
   // catColumnsIn is a collection if columnnames of the catalog
   // opt. catColumnsOut is the application equivalent of catColumnsIn
   // catColumnsOut copied from catColumnsIn, if NULL
   // !!! catColumnsIn, catColumnsOut must match in their meanings !!!
   //
   // extract data from VOTABLE and write plain text
   // the first line is a JSON style header

   //
   if (catColumnsOut == null) catColumnsOut = catColumnsIn;
   var temp = File.systemTempDirectory + "/votable_intermediate.txt";
   if (File.exists(temp)) File.remove(temp);

   if (File.exists(fOut)) File.remove(fOut);

   var tempFile = new File();
   tempFile.openOrCreate(temp);

   var rdr = new reader(fIn);

   rdr.open();

   var readInfos  = true;
   var readData   = false;
   var readRecord = false;    // read from <TR> .. </TR>
   var fields     = [];
   var magIndex   = -1;
   var record     = null;
   var items      = [];
   var numRecords = 0;
   var minMag     = Number.MAX_VALUE;
   var maxMag     = Number.MIN_VALUE;
   var ra         = 0;
   var dec        = 0;
   var radius     = 0;
   var source     = '';
   var colIndexes = [];
   var index      = -1;    //
   var colError   = false;

   Console.writeln("wxtract data from VOTABLE, gmag = " + magName);

   // translate magnitude name

   //if (magName != '')
   //{
   //   var i = catColumnsOut.indexOf(magName);
   //   if (i < 0) catColumnsIn.indexOf(magName);
   //   magName = catColumnsIn[i];
   //}

   if (rdr.isOpen())
   {
      while (!rdr.eof())
      {
         var line = rdr.readLn();
         //
         // read fields or data
         //
         if (readInfos)
         {
            var n = line.indexOf("<INFO");
            if (n > -1)
            {
               var name = getQValue("name", line);
               var value = getQValue("value", line);
               if (name == "Query")
               {
                  ra = getDValue("RA", line);
                  dec = getDValue("DEC", line);
                  radius = getDValue("SR", line);
               }
               else if (name == "queryParameters")
               {
                  for (var n = 0; n < value; n++)
                  {
                     line = rdr.readLn();
                     while (line.indexOf('\t') !== -1) line = line.replace('\t', '');
                     var kv = line.split('=');
                     if (kv.length == 2)
                     {
                        if (kv[0] == "-source") source = kv[1];
                        if (kv[0] == "-c.ra")   ra = kv[1].toFloat();
                        if (kv[0] == "-c.dec")  dec = kv[1].toFloat();
                     }
                  }
               }
               else if (name == "Provider")
               {
                  source = value;
               }
            }

            if (line.indexOf("<TABLEDATA>") > -1)
            {
               readInfos = false;
               readData   = true;
               //
               // check columns available
               //
               var votColumns = [];
               for (var i in fields) votColumns.push(fields[i].name);
               for (var i in catColumnsIn)
               {
                  if (votColumns.indexOf(catColumnsIn[i]) < 0)
                  {
                     colError = true;
                     Console.writeln("Columns " + catColumnsIn[i] + " not found");
                     break;
                  }
                  else colIndexes.push(votColumns.indexOf(catColumnsIn[i]));
               }
               continue;
            }
            var j = line.indexOf("<FIELD ");
            if (j > -1)
            {
               var kv = [];               // key, value objects
               var fld = new Object();
               var a = line.split(' ');
               for (var k in a)
               {
                  var b = a[k].split('=');
                  if (b.length == 2)
                  {
                     while (b[1].indexOf('"') !== -1) b[1] = b[1].replace('"', '');
                     fld[b[0]] = b[1];
                  }
               }
               if (fld.name == magName) magIndex = fields.length;
               Console.writeln("Magnitude index field#: " + magName + ", " + magIndex);
               fields.push(fld);
            }
         }
         else if (readData)
         {
            if (line.indexOf("</TABLEDATA>") > -1)
            {
               readData = false;
               break;
            }
            // *****************************************************************
            // collect the record items
            // *****************************************************************

            if (readRecord)      // continuation of record over mult lines
            {
               index = 0;
            }
            else
            {
               index = line.indexOf("<TR>");

               if (index > -1)
               {
                  readRecord = true;
                  items = [];
                  index += 4;
               }
            }

            if (readRecord)
            {
               while (index < line.length)
               {
                  var j = line.indexOf("<TD>", index);
                  if (j > -1)
                  {
                     var k = line.indexOf("</TD>", j);
                     if (k > -1)
                     {
                        items.push(line.substring(j + 4, k));
                         index = j + 5;
                     }
                     else
                     {
                        index += 4;
                     }
                  }
                  else
                  {
                     var l = line.indexOf("</TR>", index);
                     if (l > -1)
                     {
                        readRecord = false;
                        //Console.writeln("end read record "+items.length);

                        var outItems = [];
                        for (var n in colIndexes)
                           outItems.push(items[colIndexes[n]]);

                        var outStr = outItems.join('\t');
                        tempFile.outTextLn(outStr);

                        numRecords += 1;

                        if (magIndex > -1)
                        {
                           if (isNumeric(items[magIndex]))
                           {
                              var m = items[magIndex].toFloat();
                              if (m > maxMag) maxMag = m;
                              if (m < minMag) minMag = m;
                           }
                           else
                           {
                              console.writeln("mag "+items[magIndex]);
                              break;
                           }
                        }
                        index = l + 5;
                     }
                     else
                        index = line.length;
                  }
               }
            }
         }
      }
   }

   rdr.close();
   tempFile.close();
   if (colError) return false;
   //
   // build header
   //
   var header = {"ra":ra,"dec":dec,"columns":catColumnsOut,
                 "radius":radius,"source":source,outMax:0,
                 "records":numRecords,minMag:minMag,maxMag:maxMag};

   var outFile = new File();
   outFile.openOrCreate(fOut);

   outFile.outTextLn(JSON.stringify(header));

   rdr = new reader(temp);
   rdr.open();

   if (rdr.isOpen())
   {
      while (!rdr.eof())
      {
         var str = rdr.readLn();
         if (str == '') continue;
         outFile.outTextLn(str);
      }
   }

   rdr.close();
   outFile.close();

   File.remove(temp);

   Console.writeln("##### end #####");
   return true;
}

function isNumeric(n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
}


function getQValue(key, line)
{
   // get string value in quotes
   // i.e.: ID="Query" name="Query" value="RA=187.70788735&DEC=...."
   var i = line.indexOf(key);
   if (i < 0) return '';
   i += key.length + 1;
   var j = line.indexOf('"', i + 1);
   if (j < 0) return '';
   return line.substring(i + 1, j);
}

function getDValue(key, line)
{
   // get decimal value
   // i.e.: RA=187.70788735
   var i = line.indexOf(key + '=');
   if (i < 0) return '';
   i += key.length + 1;
   var j = i;
   while (j < line.length)
   {
      if (!"+-.1234567890".contains(line.charAt(j))) break;
      j += 1;
   }
   var subStr = line.substring(i, j);
   return subStr.tryToDouble();  // return value or null
}

// *****************************************************************************
//
// VOTABLE convert and read
//
// *****************************************************************************

// *****************************************************************************
//
// text reader
//
// *****************************************************************************

function reader(filename)
{
   const maxBlock = 65536;
   this.filename  = filename;
   this.errormsg  = "";
   this.length    = 0;
   var file       = null;
   var buffer     = null;
   var position   = 0;
   var blockSize  = 0;
   var line       = '';
   var eof        = false;
   var isOpen     = false;

   this.error = function()
   {
      return this.errormsg.length > 0;
   }

   this.eof       = function()
   {
      if (!isOpen) throw ("The file is is not open");
      return eof;
   }

   this.isOpen    = function()
   {
      return isOpen;
   }

   this.open = function()
   {
      if (!File.exists)
      {
         this.errormsg = "File not found: " +this.filename;
      }
      file = new File();
      try
      {
         file.openForReading(this.filename);
      }
      catch (ex)
      {
         this.errormsg = "File open error: " + '\n' + this.filename;
      }
      if (this.error()) return false;
      this.length = file.size;
      blockSize   = maxBlock;
      readBuffer();
      line        = '';
      eof         = false;
      isOpen      = !eof;
      return true;
   }

   this.readLn = function()
   {
      if (eof) throw ("The file is closed");
      if (!isOpen) throw ("The file is is not open");

      while (true)
      {
         if (position == blockSize)
         {
            readBuffer();
            if (buffer == null)
            {
               eof = true;
               return line;      // return unterminated string
            }
         }
         // backtrace: (function () { console.writeln(new Error().stack); })();
         var i = buffer.linearSearch(10, position);
         if (i > -1)
         {
            line += buffer.toString(position, i - position);
            position = i + 1;
            var Line = line;
            line = "";
            return Line;
         }
         else
         {
            line += buffer.toString(position, blockSize - position);
            position = blockSize;
         }
      }
   }

   this.close = function()
   {
      if (file != null)
      {
         file.close();
         file = null;
      }
      buffer = null;
   }

   this.rewind = function()
   {
      if (!isOpen) throw ("The file is is not open");
      eof = false;
      file.position = 0;
      blockSize   = maxBlock;
      readBuffer();
   }

   function readBuffer()
   {
      blockSize = Math.min(blockSize, file.size - file.position);
      var intBuffer = null;
      if (blockSize > 0)
         intBuffer =file.read( DataType_Int8Array, blockSize);
      else
      {
         buffer = null;
         return;
      }
      buffer   = new ByteArray(intBuffer);
      position = 0;
   }
}

function getCatalogType(catalogName)
{
   for (var i in catalogTypes)
   {
      if (catalogTypes[i].id == catalogName) return catalogTypes[i];
   }
   return null;
}

function fieldLabel( parent, text, width )
{
   this.label = new Label( parent );
   this.label.text = text;
   this.label.textAlignment = TextAlign_Right | TextAlign_VertCenter;
   if ( width != undefined && width != null )
      this.label.setFixedWidth( width );
   return this.label;
}

// DMSangle: Helper class for simplifying the use of angles in DMS format
// ******************************************************************
function DMSangle()
{
   this.deg = 0;
   this.min = 0;
   this.sec = 0;
   this.sign = 1;

   this.GetValue = function()
   {
      return this.sign*(this.deg + this.min/60 + this.sec/3600);
   };

   this.ToString = function( hours )
   {
      var plus = hours ? "" : "+";
      if ( this.deg != null && this.min != null && this.sec != null && this.sign != null )
         return ((this.sign < 0) ? "-": plus) +
               format( "%02d %02d %0*.*f", this.deg, this.min, hours ? 6 : 5, hours ? 3 : 2, this.sec );
      return "<* invalid *>";
   };
}


DMSangle.FromString = function( coordStr, mindeg, maxdeg )
{
   var match = coordStr.match( "'?([+-]?)([0-9]*)[ :]([0-9]*)[ :]([0-9]*(.[0-9]*)?)'?" );
   if( match==null )
      return null;
   var coord = new DMSangle();
   if ( match.length < 4 )
      throw new Error( "Invalid coordinates" );
   coord.deg = parseInt( match[2], 10 );
   if ( coord.deg < mindeg || coord.deg > maxdeg )
      throw new Error( "Invalid coordinates" );
   coord.min = parseInt( match[3], 10 );
   if ( coord.min < 0 || coord.min >= 60 )
      throw new Error( "Invalid coordinates (minutes)" );
   coord.sec = parseFloat( match[4] );
   if ( coord.sec < 0 || coord.sec >= 60 )
      throw new Error( "Invalid coordinates (seconds)" );
   coord.sign = (match[1] == '-') ? -1 : 1;
   return coord;
}

DMSangle.FromAngle = function (angle)
{
   var coord = new DMSangle();
   if (angle < 0)
   {
      coord.sign = -1;
      angle = -angle;
   }
   coord.deg = Math.floor(angle);
   coord.min = Math.floor((angle - coord.deg) * 60);
   coord.sec = (angle - coord.deg - coord.min / 60) * 3600;

   if (coord.sec > 59.999)
   {
      coord.sec = 0;
      coord.min++;
      if (coord.min == 60)
      {
         coord.min = 0;
         coord.deg++;
      }
   }

   return coord;
}

//

function FluXXLocalDR3(ra, dec, radius, magnitudeLow, magnitudeHigh, maxStars, folder)
{
   var baseFile = ra.toFixed(6);
   if (dec < 0)
      baseFile += dec.toFixed(6);
   else
      baseFile += '+' + dec.toFixed(6);

   var GaiaDR3File = folder + '/' + baseFile + '.dr3';

   if (File.exists(GaiaDR3File)) File.remove( GaiaDR3File );

   Console.writeln("Run Gaia process:");

   var P = new Gaia;
   P.command = "search";
   P.centerRA = ra;
   P.centerDec = dec;
   P.radius = radius;
   P.magnitudeLow = magnitudeLow;
   P.magnitudeHigh = magnitudeHigh;
   P.sourceLimit = maxStars;
   P.requiredFlags = 0;
   P.inclusionFlags = 0;
   P.exclusionFlags = 0;
   P.verbosity = 2;
   P.dataRelease = Gaia.prototype.DataRelease_E3;
   P.sortBy = Gaia.prototype.SortBy_G;
   P.generateTextOutput = false;
   P.textFormat = Gaia.prototype.TextHeaders_None ;
   P.textHeaders = Gaia.prototype.TextHeaders_TableColumns ;
   //P.outputFilePath = "";

   P.executeGlobal();

   Console.writeln('Gaia Process done');
   Console.flush();

   var sources = P.sources;

   var objects = [];

   for (var i = 0; i < sources.length; i++)
   {
      var s = sources[i];
      var ora = s[0];
      var odec = s[1];
      var omagG = s[5];
      var omagBP = s[6];
      var omagRP = s[7];

      if (omagBP < 1 | omagRP < 1) continue;

      var obj = [];
      obj.push(ora);
      obj.push(odec);
      obj.push(omagG);
      obj.push(omagBP);
      obj.push(omagRP);
      objects.push(obj.join('\t'));
   }

   Console.writeln('Selected sources ........... ' + objects.length);
   Console.flush();
   //
   // build header
   //
   var dataColumns = ["RAJ2000","DEJ2000","Gmag","BPmag","RPmag"];
   var header = {"ra":ra,"dec":dec, "columns":dataColumns,
                 "radius":radius, "source":"LocalCat", outMax:0,
                 "records":objects.length - 1, minMag:magnitudeLow,maxMag:magnitudeHigh};

   var outfile = new File(GaiaDR3File, FileMode_Create );
   outfile.outTextLn(JSON.stringify(header));
   //
   // write all stars to file
   //
   for (var i = 0; i < objects.length; i++) outfile.outTextLn(objects[i]);

   outfile.close();

   Console.writeln(objects.length + ' Objects found');
   Console.flush();

   return GaiaDR3File;
}


showDialog.prototype = new Dialog;
infoDialog.prototype = new Dialog;


function main()
{
   // *******************************************************************************
   //
   // check view on color and WCS
   //
   if ( CoreApplication === undefined ||
     CoreApplication.versionRevision === undefined ||
     CoreApplication.versionMajor*1e11
   + CoreApplication.versionMinor*1e8
   + CoreApplication.versionRelease*1e5
   + CoreApplication.versionRevision*1e2 < 100800800700 )
   {
      throw new Error( "This script requires PixInsight core version 1.8.8-7 or higher." );
   }

   var window = ImageWindow.activeWindow;

   if ( window.isNull )
   {
      errMessage( "No active image" );
      return;
   }
	if (!window.currentView.image.isColor)
   {
      errMessage( window.currentView.id + " is not a color image" );
      return;
   }
   else
   {
      let summary = window.astrometricSolutionSummary();
      if ( summary.length == 0 )
      {
         var txt =  "The image has no WCS coordinates. Run ImageSolver script and start FluXX again" ;
         console.writeln(txt);
         var bx = new MessageBox(txt, "Information");
         bx.execute();
         return;
      }
   }
   // *******************************************************************************
   // run dialog
   // *******************************************************************************
   //Console.hide();
   Console.show();
 	var dialog = new showDialog(window);
	dialog.execute();
}

main();
